import os

import requests  # Used for making HTTP requests
from bs4 import BeautifulSoup
from markdown_it import MarkdownIt

# Define constants for the script
CHUNK_SIZE = 1024  # Size of chunks to read/write at a time
ELEVEN_LABS_API_KEY = os.getenv("ELEVEN_LABS_API_KEY", None)  # Your API key for authentication
VOICE_ID = "RPEIZnKMqlQiZyZd1Dae"  # ID of the voice model to use. MALE - Christopher - friendly guy next door.
ELEVEN_API_URL = "https://api.elevenlabs.io/v1/text-to-speech"  # Base URL for the Text-to-Speech API

markdown_renderer = MarkdownIt()


def is_eleven_labs_enabled():
    return ELEVEN_LABS_API_KEY is not None


def generate_text_to_speech(
    text_to_speak: str,
    voice_id: str = VOICE_ID,
):
    if not is_eleven_labs_enabled():
        return "Eleven Labs API key is not set"

    # Convert the incoming text from markdown format to plain text
    html = markdown_renderer.render(text_to_speak)
    text = "".join(BeautifulSoup(html, features="lxml").findAll(text=True))

    # Construct the URL for the Text-to-Speech API request
    tts_url = f"{ELEVEN_API_URL}/{voice_id}/stream"

    # Set up headers for the API request, including the API key for authentication
    headers = {"Accept": "application/json", "xi-api-key": ELEVEN_LABS_API_KEY}

    # Set up the data payload for the API request, including the text and voice settings
    data = {
        "text": text,
        # "model_id": "eleven_multilingual_v2",
        "voice_settings": {"stability": 0.5, "similarity_boost": 0.8, "style": 0.0, "use_speaker_boost": True},
    }

    # Make the POST request to the TTS API with headers and data, enabling streaming response
    response = requests.post(tts_url, headers=headers, json=data, stream=True)

    if response.ok:
        return response
    else:
        raise Exception(f"Failed to generate text-to-speech: {response.text}")



import argparse
import logging
import pathlib
from importlib.metadata import version

logger = logging.getLogger(__name__)


def cli(args=None):
    # Setup Argument Parser for the Commandline Interface
    parser = argparse.ArgumentParser(description="Start Khoj; An AI personal assistant for your Digital Brain")
    parser.add_argument(
        "--log-file",
        default="~/.khoj/khoj.log",
        type=pathlib.Path,
        help="File path for server logs. Default: ~/.khoj/khoj.log",
    )
    parser.add_argument("--verbose", "-v", action="count", default=0, help="Show verbose conversion logs. Default: 0")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="Host address of the server. Default: 127.0.0.1")
    parser.add_argument("--port", "-p", type=int, default=42110, help="Port of the server. Default: 42110")
    parser.add_argument(
        "--socket",
        type=pathlib.Path,
        help="Path to UNIX socket for server. Use to run server behind reverse proxy. Default: /tmp/uvicorn.sock",
    )
    parser.add_argument("--sslcert", type=str, help="Path to SSL certificate file")
    parser.add_argument("--sslkey", type=str, help="Path to SSL key file")
    parser.add_argument("--version", "-V", action="store_true", help="Print the installed Khoj version and exit")
    parser.add_argument(
        "--anonymous-mode",
        action="store_true",
        default=False,
        help="Run Khoj in single user mode with no login required. Useful for personal use or testing.",
    )
    parser.add_argument(
        "--non-interactive",
        action="store_true",
        default=False,
        help="Start Khoj in non-interactive mode. Assumes interactive shell unavailable for config. E.g when run via Docker.",
    )

    args, remaining_args = parser.parse_known_args(args)

    if len(remaining_args) > 0:
        logger.info(f"⚠️  Ignoring unknown commandline args: {remaining_args}")

    args.version_no = version("khoj")
    if args.version:
        # Show version of khoj installed and exit
        print(args.version_no)
        exit(0)

    return args



# System Packages
from __future__ import annotations  # to avoid quoting type hints

from enum import Enum


class SearchType(str, Enum):
    All = "all"
    Org = "org"
    Markdown = "markdown"
    Image = "image"
    Pdf = "pdf"
    Github = "github"
    Notion = "notion"
    Plaintext = "plaintext"
    Docx = "docx"



import json
import logging
from copy import deepcopy
from datetime import datetime
from textwrap import dedent
from typing import List, Optional, cast

from openai.types.responses import Response, ResponseOutputItem

from khoj.database.models import ChatModel
from khoj.processor.conversation.utils import AgentMessage
from khoj.processor.operator.operator_actions import (
    BackAction,
    ClickAction,
    DoubleClickAction,
    DragAction,
    GotoAction,
    KeypressAction,
    MoveAction,
    NoopAction,
    OperatorAction,
    Point,
    RequestUserAction,
    ScreenshotAction,
    ScrollAction,
    TypeAction,
    WaitAction,
)
from khoj.processor.operator.operator_agent_base import AgentActResult, OperatorAgent
from khoj.processor.operator.operator_environment_base import (
    EnvironmentType,
    EnvState,
    EnvStepResult,
)
from khoj.utils.helpers import get_openai_async_client, is_none_or_empty

logger = logging.getLogger(__name__)


# --- Anthropic Operator Agent ---
class OpenAIOperatorAgent(OperatorAgent):
    async def act(self, current_state: EnvState) -> AgentActResult:
        safety_check_prefix = "Say 'continue' after resolving the following safety checks to proceed:"
        safety_check_message = None
        actions: List[OperatorAction] = []
        action_results: List[dict] = []
        self._commit_trace()  # Commit trace before next action
        system_prompt = self.get_instructions(self.environment_type, current_state)
        tools = self.get_tools(self.environment_type, current_state)
        if is_none_or_empty(self.messages):
            self.messages = [AgentMessage(role="user", content=self.query)]

        response = await self._call_model(self.vision_model, system_prompt, tools)
        self.messages += [AgentMessage(role="assistant", content=response.output)]
        rendered_response = self._render_response(response.output, current_state.screenshot)

        last_call_id = None
        content = None
        for block in response.output:
            action_to_run: Optional[OperatorAction] = None
            if block.type == "function_call":
                last_call_id = block.call_id
                if block.name == "goto":
                    try:
                        args = json.loads(block.arguments)
                        url = args.get("url")
                        if url:
                            action_to_run = GotoAction(url=url)
                        else:
                            logger.warning("Goto function called without URL argument.")
                    except json.JSONDecodeError:
                        logger.warning(f"Failed to parse arguments for goto: {block.arguments}")
                elif block.name == "back":
                    action_to_run = BackAction()

            elif block.type == "computer_call":
                last_call_id = block.call_id
                if block.pending_safety_checks:
                    safety_check_body = "\n- ".join([check.message for check in block.pending_safety_checks])
                    safety_check_message = f"{safety_check_prefix}\n- {safety_check_body}"
                    action_to_run = RequestUserAction(request=safety_check_message)
                    actions.append(action_to_run)
                    break  # Stop processing actions if safety check needed

                # Convert OpenAI action to standardized BrowserAction
                openai_action = block.action
                action_type = openai_action.type
                try:
                    if action_type == "click":
                        action_to_run = ClickAction(x=openai_action.x, y=openai_action.y, button=openai_action.button)
                    elif action_type == "double_click":
                        action_to_run = DoubleClickAction(x=openai_action.x, y=openai_action.y)
                    elif action_type == "scroll":
                        action_to_run = ScrollAction(
                            x=openai_action.x,
                            y=openai_action.y,
                            scroll_x=openai_action.scroll_x,
                            scroll_y=openai_action.scroll_y,
                        )
                    elif action_type == "keypress":
                        action_to_run = KeypressAction(keys=openai_action.keys)
                    elif action_type == "type":
                        action_to_run = TypeAction(text=openai_action.text)
                    elif action_type == "wait":
                        action_to_run = WaitAction(duration=2.0)  # OpenAI doesn't specify duration, default?
                    elif action_type == "screenshot":
                        action_to_run = ScreenshotAction()
                    elif action_type == "move":
                        action_to_run = MoveAction(x=openai_action.x, y=openai_action.y)
                    elif action_type == "drag":
                        action_to_run = DragAction(path=[Point(x=p.x, y=p.y) for p in openai_action.path])
                    else:
                        raise ValueError(f"Unsupported OpenAI computer action type: {action_type}")
                except ValueError as ve:
                    logger.error(f"Error converting OpenAI action {action_type}: {ve}")
                    content = f"ValueError: {action_type}: {ve}"
                except Exception as e:
                    logger.error(f"Error converting OpenAI action {action_type}: {e}")
                    content = f"Error: {action_type}: {e}"
            elif block.type == "message":
                rendered_response["text"] = response.output_text
            elif block.type == "reasoning":
                actions.append(NoopAction())
                # Add placeholder action result for reasoning
                # This is to prevent run termination.
                # It will be removed later by add_action_results func
                action_results.append(
                    {
                        "type": block.type,
                        "id": block.id,
                        "summary": [],
                    }
                )
            else:
                logger.warning(f"Unsupported response block type: {block.type}")
                content = f"Unsupported response block type: {block.type}"
            if action_to_run or content:
                actions.append(action_to_run)
            if action_to_run or content:
                # Prepare the action result
                action_results.append(
                    {
                        "type": f"{block.type}_output",
                        "output": content,  # Updated after environment step
                        "call_id": last_call_id,
                    }
                )

        self._update_usage(response.usage.input_tokens, response.usage.output_tokens)

        return AgentActResult(
            actions=actions,
            action_results=action_results,
            rendered_response=rendered_response,
        )

    def add_action_results(self, env_steps: list[EnvStepResult], agent_action: AgentActResult) -> None:
        if not agent_action.action_results:
            return

        # Update action results with results of applying suggested actions on the environment
        items_to_pop = []
        for idx, env_step in enumerate(env_steps):
            action_result = agent_action.action_results[idx]
            result_content = env_step.error or env_step.output or "[Action completed]"
            if env_step.type == "image" and isinstance(result_content, dict):
                # Add screenshot data in openai message format
                action_result["output"] = {
                    "type": "input_image",
                    "image_url": f"data:image/webp;base64,{result_content['image']}",
                    "current_url": result_content["url"],
                }
            elif action_result["type"] == "computer_call_output" and idx == len(env_steps) - 1:
                # Always add screenshot, current url to last action result, when computer tool used
                action_result["output"] = {
                    "type": "input_image",
                    "image_url": f"data:image/webp;base64,{env_step.screenshot_base64}",
                    "current_url": env_step.current_url,
                }
            elif action_result["type"] == "reasoning":
                items_to_pop.append(idx)  # Mark placeholder reasoning action result for removal
                continue
            elif action_result["type"] == "computer_call" and action_result["status"] == "in_progress":
                if isinstance(result_content, dict):
                    result_content["status"] = "completed"  # Mark in-progress actions as completed
                action_result["output"] = result_content
            else:
                # Add text data
                action_result["output"] = result_content

        for idx in reversed(items_to_pop):
            agent_action.action_results.pop(idx)

        self.messages += [AgentMessage(role="environment", content=agent_action.action_results)]

    async def summarize(self, current_state: EnvState, summarize_prompt: str = None) -> str:
        summarize_prompt = summarize_prompt or self.summarize_prompt
        self.messages.append(AgentMessage(role="user", content=summarize_prompt))
        response = await self._call_model(self.vision_model, summarize_prompt, [])
        self.messages += [AgentMessage(role="assistant", content=response.output)]
        if not self.messages:
            return "No actions to summarize."
        return self._compile_response(self.messages[-1].content)

    async def _call_model(self, model: ChatModel, system_prompt, tools) -> Response:
        client = get_openai_async_client(model.ai_model_api.api_key, model.ai_model_api.api_base_url)
        if tools:
            model_name = "computer-use-preview"
        else:
            model_name = model.name

        # Format messages for OpenAI API
        messages_for_api = self._format_message_for_api(self.messages)
        # format messages for summary if model is not computer-use-preview
        if model_name != "computer-use-preview":
            messages_for_api = self._format_messages_for_summary(messages_for_api)

        response: Response = await client.responses.create(
            model=model_name,
            input=messages_for_api,
            instructions=system_prompt,
            tools=tools,
            parallel_tool_calls=False,
            truncation="auto",
        )

        logger.debug(f"Openai response: {response.model_dump_json()}")
        return response

    def _format_message_for_api(self, messages: list[AgentMessage]) -> list:
        """Format the message for OpenAI API."""
        formatted_messages: list = []
        for message in messages:
            if message.role == "assistant":
                if isinstance(message.content, list):
                    # Remove reasoning message if not followed by computer call
                    if (
                        len(message.content) > 1
                        and all(hasattr(item, "type") for item in message.content)
                        and message.content[0].type == "reasoning"
                        and message.content[1].type != "computer_call"
                    ) or (
                        len(message.content) == 1
                        and all(hasattr(item, "type") for item in message.content)
                        and message.content[0].type == "reasoning"
                    ):
                        logger.warning(
                            f"Removing reasoning message not followed by a computer call action: {message.content}"
                        )
                        message.content.pop(0)
                    formatted_messages.extend(message.content)
                else:
                    logger.warning(f"Expected message content list from assistant, got {type(message.content)}")
            elif message.role == "environment":
                formatted_messages.extend(message.content)
            else:
                if isinstance(message.content, list):
                    message.content = "\n".join([part["text"] for part in message.content if part["type"] == "text"])
                formatted_messages.append(
                    {
                        "role": message.role,
                        "content": message.content,
                    }
                )

        return formatted_messages

    def _compile_response(self, response_content: str | list[dict | ResponseOutputItem]) -> str:
        """Compile the response from model into a single string for prompt tracing."""
        # Handle case where response content is a string.
        # This is the case when response content is a user query
        if isinstance(response_content, str):
            return response_content
        elif is_none_or_empty(response_content):
            return ""
        # Handle case where response_content is a dictionary and not ResponseOutputItem
        # This is the case when response_content contains action results
        if not hasattr(response_content[0], "type"):
            return "**Action**: " + json.dumps(response_content[0].get("output", "Noop"))

        compiled_response = [""]
        for block in deepcopy(response_content):
            block = cast(ResponseOutputItem, block)  # Ensure block is of type ResponseOutputItem
            # Handle different block types
            if block.type == "message":
                # Extract text content if available
                for content in block.content:
                    text_content = ""
                    if hasattr(content, "text"):
                        text_content += content.text
                    elif hasattr(content, "refusal"):
                        text_content += f"Refusal: {content.refusal}"
                    else:
                        text_content += content.model_dump_json()
                compiled_response.append(text_content)
            elif block.type == "function_call":
                block_function_input = {"action": block.name}
                if block.name == "goto":
                    try:
                        args = json.loads(block.arguments)
                        block_function_input["url"] = args.get("url", "[Missing URL]")
                    except json.JSONDecodeError:
                        block_function_input["arguments"] = block.arguments  # Show raw args on error
                compiled_response.append(f"**Action**: {json.dumps(block_function_input)}")
            elif block.type == "computer_call":
                block_computer_input = block.action
                # If it's a screenshot action
                if block_computer_input.type == "screenshot":
                    # Use a placeholder for screenshot data
                    block_input_render = block_computer_input.model_dump()
                    block_input_render["image"] = "[placeholder for screenshot]"
                    compiled_response.append(f"**Action**: {json.dumps(block_input_render)}")
                else:
                    compiled_response.append(f"**Action**: {block_computer_input.model_dump_json()}")
            elif block.type == "reasoning" and block.summary:
                compiled_response.append(f"**Thought**: {block.summary}")
        return "\n- ".join(filter(None, compiled_response))  # Filter out empty strings

    def _render_response(self, response_content: list[ResponseOutputItem], screenshot: str | None) -> dict:
        """Render OpenAI response for display, potentially including screenshots."""
        render_texts = []
        for block in deepcopy(response_content):  # Use deepcopy to avoid modifying original
            if block.type == "message":
                text_content = block.text if hasattr(block, "text") else block.model_dump_json()
                render_texts += [text_content]
            elif block.type == "function_call":
                if block.name == "goto":
                    args = json.loads(block.arguments)
                    render_texts = [f"Open URL: {args.get('url', '[Missing URL]')}"]
                else:
                    render_texts += [block.name]
            elif block.type == "computer_call":
                block_input = block.action
                if block_input.type == "screenshot" and not screenshot:
                    render_texts += ["Failed to get screenshot"]
                elif block_input.type == "type":
                    render_texts += [f'Type "{block_input.text}"']
                elif block_input.type == "keypress":
                    render_texts += [f"Press {'+'.join(block_input.keys)}"]
                else:
                    render_texts += [f"{block_input.type.capitalize()}"]
            elif block.type == "reasoning" and block.summary:
                render_texts += [f"**Thought**: {block.summary}"]

        render_payload = {
            # Combine text into a single string and filter out empty strings
            "text": "\n- ".join(filter(None, render_texts)),
            # Add screenshot data if available
            "image": f"data:image/webp;base64,{screenshot}" if screenshot else None,
        }

        return render_payload

    def get_instructions(self, environment_type: EnvironmentType, current_state: EnvState) -> str:
        """Return system instructions for the OpenAI operator."""
        if environment_type == EnvironmentType.BROWSER:
            return dedent(
                f"""
                <SYSTEM_CAPABILITY>
                * You are Khoj, a smart web browser operating assistant. You help the users accomplish tasks using a web browser.
                * You operate a single Chromium browser page using Playwright.
                * You cannot access the OS or filesystem.
                * You can interact with the web browser to perform tasks like clicking, typing, scrolling, and more using the computer_use_preview tool.
                * You can use the additional back() and goto() functions to navigate the browser.
                * Always use the goto() function to navigate to a specific URL. If you see nothing, try goto duckduckgo.com
                * When viewing a webpage it can be helpful to zoom out so that you can see everything on the page. Either that, or make sure you scroll down to see everything before deciding something isn't available.
                * When using your computer function calls, they take a while to run and send back to you. Where possible/feasible, try to chain multiple of these calls all into one function calls request.
                * Perform web searches using DuckDuckGo. Don't use Google even if requested as the query will fail.
                * The current date is {datetime.today().strftime("%A, %B %-d, %Y")}.
                * The current URL is {current_state.url}.
                </SYSTEM_CAPABILITY>

                <IMPORTANT>
                * You are allowed upto {self.max_iterations} iterations to complete the task.
                * After initialization if the browser is blank, enter a website URL using the goto() function instead of waiting
                </IMPORTANT>
                """
            ).lstrip()
        elif environment_type == EnvironmentType.COMPUTER:
            return dedent(
                f"""
                <SYSTEM_CAPABILITY>
                * You are Khoj, a smart computer operating assistant. You help the users accomplish their tasks using a computer.
                * You can interact with the computer to perform tasks like clicking, typing, scrolling, and more using the computer_use_preview tool.
                * When viewing a document or webpage it can be helpful to zoom out or scroll down to ensure you see everything before deciding something isn't available.
                * When using your computer function calls, they take a while to run and send back to you. Where possible/feasible, try to chain multiple of these calls all into one function calls request.
                * Perform web searches using DuckDuckGo. Don't use Google even if requested as the query will fail.
                * You are allowed upto {self.max_iterations} iterations to complete the task.
                </SYSTEM_CAPABILITY>

                <CONTEXT>
                * The current date is {datetime.today().strftime("%A, %B %-d, %Y")}.
                </CONTEXT>
                """
            ).lstrip()
        else:
            raise ValueError(f"Unsupported environment type: {environment_type}")

    def get_tools(self, environment_type: EnvironmentType, current_state: EnvState) -> list[dict]:
        """Return the tools available for the OpenAI operator."""
        if environment_type == EnvironmentType.COMPUTER:
            # TODO: Get OS info from the environment
            # For now, assume Linux as the environment OS
            environment_os = "linux"
            # environment = "mac" if platform.system() == "Darwin" else "windows" if platform.system() == "Windows" else "linux"
        else:
            environment_os = "browser"

        tools = [
            {
                "type": "computer_use_preview",
                "display_width": current_state.width,
                "display_height": current_state.height,
                "environment": environment_os,
            }
        ]
        if environment_type == EnvironmentType.BROWSER:
            tools += [
                {
                    "type": "function",
                    "name": "back",
                    "description": "Go back to the previous page.",
                    "parameters": {},
                },
                {
                    "type": "function",
                    "name": "goto",
                    "description": "Go to a specific URL.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "url": {
                                "type": "string",
                                "description": "Fully qualified URL to navigate to.",
                            },
                        },
                        "additionalProperties": False,
                        "required": ["url"],
                    },
                },
            ]
        return tools

    def _format_messages_for_summary(self, formatted_messages: List[dict]) -> List[dict]:
        """Format messages for summary."""
        # Format messages to interact with non computer use AI models
        items_to_drop = []  # Track indices to drop reasoning messages
        for idx, msg in enumerate(formatted_messages):
            if isinstance(msg, dict) and "content" in msg:
                continue
            elif isinstance(msg, dict) and "output" in msg:
                # Drop current_url from output as not supported for non computer operations
                if "current_url" in msg["output"]:
                    del msg["output"]["current_url"]
                formatted_messages[idx] = {"role": "user", "content": [msg["output"]]}
            elif isinstance(msg, str):
                formatted_messages[idx] = {"role": "user", "content": [{"type": "input_text", "text": msg}]}
            else:
                text = self._compile_response([msg])
                if not text:
                    items_to_drop.append(idx)  # Track index to drop reasoning message
                else:
                    formatted_messages[idx] = {
                        "role": "assistant",
                        "content": [{"type": "output_text", "text": text}],
                    }

        # Remove reasoning messages for non-computer use models
        for idx in reversed(items_to_drop):
            formatted_messages.pop(idx)

        return formatted_messages



import json
import logging
from copy import deepcopy
from time import perf_counter
from typing import AsyncGenerator, Dict, List

import anthropic
from langchain_core.messages.chat import ChatMessage
from pydantic import BaseModel
from tenacity import (
    before_sleep_log,
    retry,
    stop_after_attempt,
    wait_exponential,
    wait_random_exponential,
)

from khoj.processor.conversation.utils import (
    ResponseWithThought,
    ToolCall,
    commit_conversation_trace,
    get_image_from_base64,
    get_image_from_url,
)
from khoj.utils.helpers import (
    ToolDefinition,
    create_tool_definition,
    get_anthropic_async_client,
    get_anthropic_client,
    get_chat_usage_metrics,
    is_none_or_empty,
    is_promptrace_enabled,
)

logger = logging.getLogger(__name__)

anthropic_clients: Dict[str, anthropic.Anthropic | anthropic.AnthropicVertex] = {}
anthropic_async_clients: Dict[str, anthropic.AsyncAnthropic | anthropic.AsyncAnthropicVertex] = {}

DEFAULT_MAX_TOKENS_ANTHROPIC = 8000
MAX_REASONING_TOKENS_ANTHROPIC = 12000
REASONING_MODELS = ["claude-3-7", "claude-sonnet-4", "claude-opus-4"]


@retry(
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(2),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def anthropic_completion_with_backoff(
    messages: list[ChatMessage],
    system_prompt: str,
    model_name: str,
    temperature: float = 0.4,
    api_key: str | None = None,
    api_base_url: str | None = None,
    model_kwargs: dict | None = None,
    max_tokens: int | None = None,
    response_type: str = "text",
    response_schema: BaseModel | None = None,
    tools: List[ToolDefinition] = None,
    deepthought: bool = False,
    tracer: dict = {},
) -> ResponseWithThought:
    client = anthropic_clients.get(api_key)
    if not client:
        client = get_anthropic_client(api_key, api_base_url)
        anthropic_clients[api_key] = client

    formatted_messages, system = format_messages_for_anthropic(messages, system_prompt)

    thoughts = ""
    aggregated_response = ""
    final_message = None
    model_kwargs = model_kwargs or dict()

    # Configure structured output
    if tools:
        # Convert tools to Anthropic format
        model_kwargs["tools"] = [
            anthropic.types.ToolParam(name=tool.name, description=tool.description, input_schema=tool.schema)
            for tool in tools
        ]
        # Cache tool definitions
        last_tool = model_kwargs["tools"][-1]
        last_tool["cache_control"] = {"type": "ephemeral"}
    elif response_schema:
        tool = create_tool_definition(response_schema)
        model_kwargs["tools"] = [
            anthropic.types.ToolParam(name=tool.name, description=tool.description, input_schema=tool.schema)
        ]
    elif response_type == "json_object" and not (is_reasoning_model(model_name) and deepthought):
        # Prefill model response with '{' to make it output a valid JSON object. Not supported with extended thinking.
        formatted_messages.append(anthropic.types.MessageParam(role="assistant", content="{"))
        aggregated_response += "{"

    if system:
        model_kwargs["system"] = system

    max_tokens = max_tokens or DEFAULT_MAX_TOKENS_ANTHROPIC
    if deepthought and is_reasoning_model(model_name):
        model_kwargs["thinking"] = {"type": "enabled", "budget_tokens": MAX_REASONING_TOKENS_ANTHROPIC}
        max_tokens += MAX_REASONING_TOKENS_ANTHROPIC
        # Temperature control not supported when using extended thinking
        temperature = 1.0

    with client.messages.stream(
        messages=formatted_messages,
        model=model_name,  # type: ignore
        temperature=temperature,
        timeout=20,
        max_tokens=max_tokens,
        **(model_kwargs),
    ) as stream:
        for chunk in stream:
            if chunk.type != "content_block_delta":
                continue
            if chunk.delta.type == "thinking_delta":
                thoughts += chunk.delta.thinking
            elif chunk.delta.type == "text_delta":
                aggregated_response += chunk.delta.text
        final_message = stream.get_final_message()

    # Track raw content of model response to reuse for cache hits in multi-turn chats
    raw_content = [item.model_dump() for item in final_message.content]

    # Extract all tool calls if tools are enabled
    if tools:
        tool_calls = [
            ToolCall(name=item.name, args=item.input, id=item.id).__dict__
            for item in final_message.content
            if item.type == "tool_use"
        ]
        if tool_calls:
            # If there are tool calls, aggregate thoughts and responses into thoughts
            if thoughts and aggregated_response:
                # wrap each line of thought in italics
                thoughts = "\n".join([f"*{line.strip()}*" for line in thoughts.splitlines() if line.strip()])
                thoughts = f"{thoughts}\n\n{aggregated_response}"
            else:
                thoughts = thoughts or aggregated_response
            # Json dump tool calls into aggregated response
            aggregated_response = json.dumps(tool_calls)
    # If response schema is used, return the first tool call's input
    elif response_schema:
        for item in final_message.content:
            if item.type == "tool_use":
                aggregated_response = json.dumps(item.input)
                break

    # Calculate cost of chat
    input_tokens = final_message.usage.input_tokens
    output_tokens = final_message.usage.output_tokens
    cache_read_tokens = final_message.usage.cache_read_input_tokens
    cache_write_tokens = final_message.usage.cache_creation_input_tokens
    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, cache_read_tokens, cache_write_tokens, usage=tracer.get("usage")
    )

    # Validate the response. If empty, raise an error to retry.
    if is_none_or_empty(aggregated_response):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over API. Retry if needed.")

    # Save conversation trace
    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        commit_conversation_trace(messages, aggregated_response, tracer)

    return ResponseWithThought(text=aggregated_response, thought=thoughts, raw_content=raw_content)


@retry(
    wait=wait_exponential(multiplier=1, min=4, max=10),
    stop=stop_after_attempt(2),
    before_sleep=before_sleep_log(logger, logging.WARNING),
    reraise=False,
)
async def anthropic_chat_completion_with_backoff(
    messages: list[ChatMessage],
    model_name: str | None,
    temperature: float,
    api_key: str | None,
    api_base_url: str,
    system_prompt: str,
    max_prompt_size: int | None = None,
    deepthought: bool = False,
    model_kwargs: dict | None = None,
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    client = anthropic_async_clients.get(api_key)
    if not client:
        client = get_anthropic_async_client(api_key, api_base_url)
        anthropic_async_clients[api_key] = client

    model_kwargs = model_kwargs or dict()
    max_tokens = DEFAULT_MAX_TOKENS_ANTHROPIC
    if deepthought and is_reasoning_model(model_name):
        model_kwargs["thinking"] = {"type": "enabled", "budget_tokens": MAX_REASONING_TOKENS_ANTHROPIC}
        max_tokens += MAX_REASONING_TOKENS_ANTHROPIC
        # Temperature control not supported when using extended thinking
        temperature = 1.0

    formatted_messages, system = format_messages_for_anthropic(messages, system_prompt)

    aggregated_response = ""
    response_started = False
    final_message = None
    start_time = perf_counter()
    async with client.messages.stream(
        messages=formatted_messages,
        model=model_name,  # type: ignore
        temperature=temperature,
        system=system,
        timeout=20,
        max_tokens=max_tokens,
        **model_kwargs,
    ) as stream:
        async for chunk in stream:
            # Log the time taken to start response
            if not response_started:
                response_started = True
                logger.info(f"First response took: {perf_counter() - start_time:.3f} seconds")
            if chunk.type == "message_delta":
                if chunk.delta.stop_reason == "refusal":
                    yield ResponseWithThought(
                        text="...I'm sorry, but my safety filters prevent me from assisting with this query."
                    )
                elif chunk.delta.stop_reason == "max_tokens":
                    yield ResponseWithThought(text="...I'm sorry, but I've hit my response length limit.")
                if chunk.delta.stop_reason in ["refusal", "max_tokens"]:
                    logger.warning(
                        f"LLM Response Prevented for {model_name}: {chunk.delta.stop_reason}.\n"
                        + f"Last Message by {messages[-1].role}: {messages[-1].content}"
                    )
                    break
            # Skip empty chunks
            if chunk.type != "content_block_delta":
                continue
            # Handle streamed response chunk
            response_chunk: ResponseWithThought = None
            if chunk.delta.type == "text_delta":
                response_chunk = ResponseWithThought(text=chunk.delta.text)
                aggregated_response += chunk.delta.text
            if chunk.delta.type == "thinking_delta":
                response_chunk = ResponseWithThought(thought=chunk.delta.thinking)
            # Handle streamed response chunk
            if response_chunk:
                yield response_chunk
        final_message = await stream.get_final_message()

    # Calculate cost of chat
    input_tokens = final_message.usage.input_tokens
    output_tokens = final_message.usage.output_tokens
    cache_read_tokens = final_message.usage.cache_read_input_tokens
    cache_write_tokens = final_message.usage.cache_creation_input_tokens
    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, cache_read_tokens, cache_write_tokens, usage=tracer.get("usage")
    )

    # Validate the response. If empty, raise an error to retry.
    if is_none_or_empty(aggregated_response):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over API. Retry if needed.")

    # Log the time taken to stream the entire response
    logger.info(f"Chat streaming took: {perf_counter() - start_time:.3f} seconds")

    # Save conversation trace
    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        commit_conversation_trace(messages, aggregated_response, tracer)


def format_messages_for_anthropic(raw_messages: list[ChatMessage], system_prompt: str = None):
    """
    Format messages for Anthropic
    """
    # Extract system prompt
    system_prompt = system_prompt or ""
    messages = deepcopy(raw_messages)
    for message in messages:
        if message.role == "system":
            if isinstance(message.content, list):
                system_prompt += "\n".join([part["text"] for part in message.content if part["type"] == "text"])
            else:
                system_prompt += message.content
            messages.remove(message)
    if not is_none_or_empty(system_prompt):
        system = [{"type": "text", "text": system_prompt, "cache_control": {"type": "ephemeral"}}]
    else:
        system = None

    # Anthropic requires the first message to be a user message unless its a tool call
    message_type = messages[0].additional_kwargs.get("message_type", None)
    if len(messages) == 1 and message_type != "tool_call":
        messages[0].role = "user"

    for message in reversed(messages):  # Process in reverse to not mess up iterator when drop invalid messages
        # Handle tool call and tool result message types from additional_kwargs
        message_type = message.additional_kwargs.get("message_type")
        if message_type == "tool_call":
            pass
        elif message_type == "tool_result":
            # Convert tool_result to Anthropic tool_result format
            content = []
            for part in message.content:
                # Skip tool results without valid tool_use_id as Anthropic API requires string IDs
                if not part.get("id"):
                    logger.warning(f"Dropping tool result without valid tool_use_id: {part.get('name')}")
                    continue
                content.append(
                    {
                        "type": "tool_result",
                        "tool_use_id": part["id"],
                        "content": part["content"],
                    }
                )
            message.content = content
        # Convert image urls to base64 encoded images in Anthropic message format
        elif isinstance(message.content, list):
            content = []
            # Sort the content. Anthropic models prefer that text comes after images.
            message.content.sort(key=lambda x: 0 if x["type"] == "image_url" else 1)
            for idx, part in enumerate(message.content):
                if part["type"] == "text":
                    content.append({"type": "text", "text": part["text"]})
                elif part["type"] == "image_url":
                    image_data = part["image_url"]["url"]
                    if image_data.startswith("http"):
                        image = get_image_from_url(image_data, type="b64")
                    else:
                        image = get_image_from_base64(image_data, type="b64")
                    # Prefix each image with text block enumerating the image number
                    # This helps the model reference the image in its response. Recommended by Anthropic
                    content.extend(
                        [
                            {
                                "type": "text",
                                "text": f"Image {idx + 1}:",
                            },
                            {
                                "type": "image",
                                "source": {"type": "base64", "media_type": image.type, "data": image.content},
                            },
                        ]
                    )
            message.content = content

        if is_none_or_empty(message.content):
            logger.warning(f"Drop message with empty content as not supported:\n{message}")
            messages.remove(message)
            continue
        if isinstance(message.content, str):
            message.content = [{"type": "text", "text": message.content}]

    # Add cache control to enable prompt caching for conversations with sufficient context
    # Only add caching if we have multiple messages to make it worthwhile
    if len(messages) > 2:
        # Remove any existing cache controls from previous messages
        for message in messages:  # All except the last message
            if isinstance(message.content, list):
                for block in message.content:
                    if isinstance(block, dict) and "cache_control" in block:
                        del block["cache_control"]

        # Add cache control to the last content block of last message.
        # Caching should improve research efficiency.
        cache_message = messages[-1]
        if isinstance(cache_message.content, list) and cache_message.content:
            # Add cache control to the last content block only if it's a text block with non-empty content
            last_block = cache_message.content[-1]
            if isinstance(last_block, dict) and (
                (last_block.get("type") == "text" and last_block.get("text", "").strip())
                or (last_block.get("type") == "tool_result" and last_block.get("content", []))
            ):
                last_block["cache_control"] = {"type": "ephemeral"}

    formatted_messages: List[anthropic.types.MessageParam] = [
        anthropic.types.MessageParam(role=message.role, content=message.content) for message in messages
    ]

    return formatted_messages, system


def is_reasoning_model(model_name: str) -> bool:
    return any(model_name.startswith(model) for model in REASONING_MODELS)



import ast
import asyncio
import base64
import io
import logging
import platform
import subprocess
from typing import Literal, Optional, Union

from PIL import Image, ImageDraw

from khoj.processor.operator.operator_actions import DragAction, OperatorAction, Point
from khoj.processor.operator.operator_environment_base import (
    Environment,
    EnvState,
    EnvStepResult,
)
from khoj.utils.helpers import convert_image_to_webp

logger = logging.getLogger(__name__)


# --- Concrete Computer Environment ---
class ComputerEnvironment(Environment):
    def __init__(
        self,
        provider: Literal["local", "docker"] = "local",
        docker_display: str = ":99",
        docker_container_name: str = "khoj-computer",
    ):
        self.provider = provider
        self.docker_display = docker_display
        self.docker_container_name = docker_container_name

        self.width: int = 0
        self.height: int = 0
        self.mouse_pos: Point = Point(x=0, y=0)

    async def _execute(self, func_name, *args, **kwargs):
        """
        Executes a pyautogui function, abstracting the execution context.
        Currently runs locally using asyncio.to_thread.
        """
        python_command_str = self.generate_pyautogui_command(func_name, *args, **kwargs)
        # Docker execution
        if self.provider == "docker":
            try:
                output_str = await self.docker_execute(python_command_str)
            except RuntimeError as e:  # Catch other Docker execution errors
                logger.error(f"Error during Docker execution of {func_name}: {e}")
                raise  # Re-raise as a general error for the caller to handle
        # Local execution
        else:
            process = await asyncio.to_thread(
                subprocess.run,
                ["python3", "-c", python_command_str],
                capture_output=True,
                text=True,
                check=False,  # We check returncode manually
            )
            output_str = process.stdout.strip()
            if process.returncode != 0:
                if "FailSafeException" in process.stderr or "FailSafeException" in process.stdout:
                    # Extract the message if possible, otherwise use generic
                    fs_msg = process.stderr or process.stdout
                    raise KeyboardInterrupt(fs_msg)
                else:
                    error_msg = (
                        f'Local script execution failed:\nCmd: python3 -c "{python_command_str[:200]}...{python_command_str[-200:]}\n'
                        f"Return Code: {process.returncode}\nStderr: {process.stderr}\nStdout: {process.stdout}"
                    )
                    logger.error(error_msg)
                    raise RuntimeError(f"Local script execution error: {process.stderr or process.stdout}")
        if not output_str or output_str == "None":
            return None

        try:
            return ast.literal_eval(output_str)
        except (ValueError, SyntaxError):
            # If not a literal (e.g., some other string output), return as is
            return output_str

    async def start(self, width: int, height: int) -> None:
        """
        Initializes the computer environment.
        The width and height parameters are logged, but actual screen dimensions are used.
        """
        screen_width, screen_height = await self._execute("size")

        self.width = screen_width
        self.height = screen_height
        # Initialize mouse position to center, or current if available
        try:
            current_x, current_y = await self._execute("position")
            self.mouse_pos = Point(x=current_x, y=current_y)
        except Exception:  # Fallback if position cannot be obtained initially
            self.mouse_pos = Point(x=self.width / 2, y=self.height / 2)

        logger.info(
            f"Computer environment started. Screen size: {self.width}x{self.height}. "
            f"Input width/height ({width}x{height}) are noted but screen dimensioning uses actual screen size. "
            f"Initial mouse position: ({self.mouse_pos.x},{self.mouse_pos.y})"
        )

    async def _get_screenshot(self) -> Optional[str]:
        try:
            # Get screenshot
            base64_png_str = await self._execute("screenshot")
            screenshot_bytes = base64.b64decode(base64_png_str)

            # Get current mouse position
            current_mouse_x, current_mouse_y = await self._execute("position")
            draw_pos = Point(x=current_mouse_x, y=current_mouse_y)

            # Add mouse position to screenshot
            screenshot_bytes_with_mouse = await self._draw_mouse_position(screenshot_bytes, draw_pos)
            screenshot_webp_bytes = convert_image_to_webp(screenshot_bytes_with_mouse)
            return base64.b64encode(screenshot_webp_bytes).decode("utf-8")
        except KeyboardInterrupt:  # Propagate keyboard interrupts
            raise
        except Exception as e:
            logger.error(f"Failed to get screenshot: {e}", exc_info=True)
            return None

    async def _draw_mouse_position(self, screenshot_bytes: bytes, mouse_pos: Point) -> bytes:
        if Image is None or ImageDraw is None:
            return screenshot_bytes
        try:
            image = Image.open(io.BytesIO(screenshot_bytes))
            draw = ImageDraw.Draw(image)
            radius = 8
            # Red circle with black border for better visibility
            draw.ellipse(
                (mouse_pos.x - radius, mouse_pos.y - radius, mouse_pos.x + radius, mouse_pos.y + radius),
                outline="black",
                fill="red",
                width=2,
            )
            output_buffer = io.BytesIO()
            image.save(output_buffer, format="PNG")
            return output_buffer.getvalue()
        except Exception as e:
            logger.error(f"Failed to draw mouse position: {e}")
            return screenshot_bytes

    async def get_state(self) -> EnvState:
        screenshot = await self._get_screenshot()
        return EnvState(screenshot=screenshot, height=self.height, width=self.width)

    async def step(self, action: OperatorAction) -> EnvStepResult:
        output: Optional[Union[str, dict]] = None
        error: Optional[str] = None
        step_type: str = "text"

        try:
            match action.type:
                case "click":
                    x, y, button_name = action.x, action.y, action.button
                    modifiers_to_press = self.parse_key_combination(action.modifiers) if action.modifiers else []
                    for mod_key in modifiers_to_press:
                        await self._execute("keyDown", mod_key)

                    if button_name == "wheel":
                        # Perform a small scroll action at this position (e.g., one "tick" down)
                        # Pyautogui scroll: positive up, negative down.
                        await self._execute("scroll", -1, x=x, y=y)
                        output = f"Scrolled wheel at ({x}, {y})"
                    else:
                        pyautogui_button = button_name.lower() if button_name else "left"
                        await self._execute("click", x=x, y=y, button=pyautogui_button)
                        output = f"{button_name.capitalize() if button_name else 'Left'} clicked at ({x}, {y})"

                    for mod_key in reversed(modifiers_to_press):
                        await self._execute("keyUp", mod_key)

                    self.mouse_pos = Point(x=x, y=y)
                    logger.debug(f"Action: {action.type} {button_name} at ({x},{y}) with modifiers {action.modifiers}")

                case "double_click":
                    x, y = action.x, action.y
                    await self._execute("doubleClick", x=x, y=y)
                    self.mouse_pos = Point(x=x, y=y)
                    output = f"Double clicked at ({x}, {y})"
                    logger.debug(f"Action: {action.type} at ({x},{y})")

                case "triple_click":
                    x, y = action.x, action.y
                    await self._execute("click", x=x, y=y, clicks=3)
                    self.mouse_pos = Point(x=x, y=y)
                    output = f"Triple clicked at ({x}, {y})"
                    logger.debug(f"Action: {action.type} at ({x},{y})")

                case "scroll":
                    current_x_pos, current_y_pos = await self._execute("position")
                    target_x = action.x if action.x is not None else current_x_pos
                    target_y = action.y if action.y is not None else current_y_pos

                    if target_x != current_x_pos or target_y != current_y_pos:
                        await self._execute("moveTo", target_x, target_y)

                    self.mouse_pos = Point(x=target_x, y=target_y)  # Update mouse pos to scroll location

                    if action.scroll_x is not None or action.scroll_y is not None:
                        scroll_x_amount = action.scroll_x or 0
                        scroll_y_amount = action.scroll_y or 0

                        if scroll_x_amount != 0:
                            await self._execute("hscroll", scroll_x_amount)
                        if scroll_y_amount != 0:
                            # pyautogui scroll: positive up, so negate for typical "scroll down" meaning positive y
                            await self._execute("scroll", -scroll_y_amount)
                        output = f"Scrolled by (x:{scroll_x_amount}, y:{scroll_y_amount}) at ({target_x}, {target_y})"
                    elif action.scroll_direction:
                        # Define scroll unit (number of pyautogui scroll 'clicks')
                        # This might need tuning based on desired sensitivity.
                        pyautogui_scroll_clicks_per_unit = 1
                        amount = action.scroll_amount or 1
                        total_scroll_clicks = pyautogui_scroll_clicks_per_unit * amount

                        if action.scroll_direction == "up":
                            await self._execute("scroll", total_scroll_clicks)
                        elif action.scroll_direction == "down":
                            await self._execute("scroll", -total_scroll_clicks)
                        elif action.scroll_direction == "left":
                            await self._execute("hscroll", -total_scroll_clicks * 3)
                        elif action.scroll_direction == "right":
                            await self._execute("hscroll", total_scroll_clicks * 3)
                        output = f"Scrolled {action.scroll_direction} by {amount} units at ({target_x}, {target_y})"
                    else:
                        error = "Scroll action requires either scroll_x/y or scroll_direction"
                    logger.debug(f"Action: {action.type} details: {output or error}")

                case "keypress":
                    mapped_keys = [self.CUA_KEY_TO_PYAUTOGUI_KEY.get(k.lower(), k.lower()) for k in action.keys]
                    key_string = "N/A"
                    if not mapped_keys:
                        error = "Keypress action requires at least one key"
                    elif len(mapped_keys) > 1:
                        await self._execute("hotkey", *mapped_keys)
                        key_string = "+".join(mapped_keys)
                    else:
                        await self._execute("press", mapped_keys[0])
                        key_string = mapped_keys[0]
                    if not error:
                        output = f"Pressed key(s): {key_string}"
                    logger.debug(f"Action: {action.type} '{key_string}'")

                case "type":
                    text_to_type = action.text
                    await self._execute("typewrite", text_to_type, interval=0.02)  # Small interval
                    output = f"Typed text: {text_to_type}"
                    logger.debug(f"Action: {action.type} '{text_to_type}'")

                case "wait":
                    duration = action.duration
                    await asyncio.sleep(duration)
                    output = f"Waited for {duration} seconds"
                    logger.debug(f"Action: {action.type} for {duration}s")

                case "screenshot":
                    step_type = "image"
                    # The actual screenshot data is added from after_state later
                    output = {"message": "Screenshot captured", "url": "desktop"}
                    logger.debug(f"Action: {action.type}")

                case "move":
                    x, y = action.x, action.y
                    await self._execute("moveTo", x, y, duration=0.2)  # Small duration for smooth move
                    self.mouse_pos = Point(x=x, y=y)
                    output = f"Moved mouse to ({x}, {y})"
                    logger.debug(f"Action: {action.type} to ({x},{y})")

                case "drag":
                    if not isinstance(action, DragAction):
                        raise TypeError("Invalid action type for drag")
                    drag_path = action.path
                    if not drag_path:
                        error = "Missing path for drag action"
                    else:
                        start_x, start_y = drag_path[0].x, drag_path[0].y
                        await self._execute("moveTo", start_x, start_y, duration=0.1)
                        await self._execute("mouseDown")
                        for point in drag_path[1:]:
                            await self._execute("moveTo", point.x, point.y, duration=0.05)
                        await self._execute("mouseUp")
                        self.mouse_pos = Point(x=drag_path[-1].x, y=drag_path[-1].y)
                        output = f"Drag along path starting at ({start_x},{start_y})"
                        logger.debug(f"Action: {action.type} with {len(drag_path)} points")

                case "mouse_down":
                    pyautogui_button = action.button.lower() if action.button else "left"
                    await self._execute("mouseDown", button=pyautogui_button)
                    output = f"{action.button.capitalize() if action.button else 'Left'} mouse button down"
                    logger.debug(f"Action: {action.type} {action.button}")

                case "mouse_up":
                    pyautogui_button = action.button.lower() if action.button else "left"
                    await self._execute("mouseUp", button=pyautogui_button)
                    output = f"{action.button.capitalize() if action.button else 'Left'} mouse button up"
                    logger.debug(f"Action: {action.type} {action.button}")

                case "hold_key":
                    keys_to_hold_str = action.text
                    duration = action.duration
                    parsed_keys = self.parse_key_combination(keys_to_hold_str)
                    if not parsed_keys:
                        error = f"No valid keys found in '{keys_to_hold_str}' for hold_key"
                    else:
                        for key_to_hold in parsed_keys:
                            await self._execute("keyDown", key_to_hold)
                        await asyncio.sleep(duration)  # Non-pyautogui, direct sleep
                        for key_to_hold in reversed(parsed_keys):  # Release in reverse order
                            await self._execute("keyUp", key_to_hold)
                        output = (
                            f"Held key{'s' if len(parsed_keys) > 1 else ''} {keys_to_hold_str} for {duration} seconds"
                        )
                        logger.debug(f"Action: {action.type} '{keys_to_hold_str}' for {duration}s")

                case "key_down":
                    key_to_press = self.CUA_KEY_TO_PYAUTOGUI_KEY.get(action.key.lower(), action.key)
                    await self._execute("keyDown", key_to_press)
                    output = f"Key down: {key_to_press}"
                    logger.debug(f"Action: {action.type} {key_to_press}")

                case "key_up":
                    key_to_release = self.CUA_KEY_TO_PYAUTOGUI_KEY.get(action.key.lower(), action.key)
                    await self._execute("keyUp", key_to_release)
                    output = f"Key up: {key_to_release}"
                    logger.debug(f"Action: {action.type} {key_to_release}")

                case "cursor_position":
                    pos_x, pos_y = await self._execute("position")
                    self.mouse_pos = Point(x=pos_x, y=pos_y)
                    output = f"Cursor position is ({pos_x}, {pos_y})"
                    logger.debug(f"Action: {action.type}, position: ({pos_x},{pos_y})")

                case "goto":
                    output = f"Goto action (URL: {action.url}) is not applicable for ComputerEnvironment."
                    logger.warning(f"Unsupported action: {action.type} for ComputerEnvironment.")

                case "back":
                    output = "Back action is not applicable for ComputerEnvironment."
                    logger.warning(f"Unsupported action: {action.type} for ComputerEnvironment.")

                case "terminal":
                    # Execute terminal command
                    result = await self._execute_shell_command(action.command)
                    if result["success"]:
                        output = f"Command executed successfully:\n{result['output']}"
                    else:
                        error = f"Command execution failed: {result['error']}"
                    logger.debug(f"Action: {action.type} with command '{action.command}'")

                case "text_editor_view":
                    # View file contents
                    file_path = action.path
                    view_range = action.view_range
                    # Type guard: path should be str for text editor actions
                    if not isinstance(file_path, str):
                        raise TypeError("Invalid path type for text editor view action")
                    escaped_path = file_path.replace("'", "'\"'\"'")
                    is_dir = await self._execute("os.path.isdir", escaped_path)
                    if is_dir:
                        cmd = rf"find {escaped_path} -maxdepth 2 -not -path '*/\.*'"
                    elif view_range:
                        # Use head/tail to view specific line range
                        start_line, end_line = view_range
                        lines_to_show = end_line - start_line + 1
                        cmd = f"head -n {end_line} '{escaped_path}' | tail -n {lines_to_show}"
                    else:
                        # View entire file
                        cmd = f"cat '{escaped_path}'"

                    result = await self._execute_shell_command(cmd)
                    MAX_OUTPUT_LENGTH = 15000  # Limit output length to avoid excessive data
                    if len(result["output"]) > MAX_OUTPUT_LENGTH:
                        result["output"] = f"{result['output'][:MAX_OUTPUT_LENGTH]}..."
                    if result["success"]:
                        if is_dir:
                            output = f"Here's the files and directories up to 2 levels deep in {file_path}, excluding hidden items:\n{result['output']}"
                        else:
                            output = f"File contents of {file_path}:\n{result['output']}"
                    else:
                        error = f"Failed to view file {file_path}: {result['error']}"
                    logger.debug(f"Action: {action.type} for file {file_path}")

                case "text_editor_create":
                    # Create new file with contents
                    file_path = action.path
                    file_text = action.file_text
                    # Type guard: path should be str for text editor actions
                    if not isinstance(file_path, str):
                        raise TypeError("Invalid path type for text editor create action")
                    escaped_path = file_path.replace("'", "'\"'\"'")
                    escaped_content = file_text.replace("\t", "    ").replace(
                        "'", "'\"'\"'"
                    )  # Escape single quotes for shell
                    cmd = f"echo '{escaped_content}' > '{escaped_path}'"

                    result = await self._execute_shell_command(cmd)
                    if result["success"]:
                        output = f"Created file {file_path} with {len(file_text)} characters"
                    else:
                        error = f"Failed to create file {file_path}: {result['error']}"
                    logger.debug(f"Action: {action.type} created file {file_path}")

                case "text_editor_str_replace":
                    # Execute string replacement
                    file_path = action.path
                    old_str = action.old_str
                    new_str = action.new_str

                    # Type guard: path should be str for text editor actions
                    if not isinstance(file_path, str):
                        raise TypeError("Invalid path type for text editor str_replace action")
                    # Use sed for string replacement, escaping special characters
                    escaped_path = file_path.replace("'", "'\"'\"'")
                    escaped_old = (
                        old_str.replace("\t", "    ")
                        .replace("\\", "\\\\")
                        .replace("\n", "\\n")
                        .replace("/", "\\/")
                        .replace("'", "'\"'\"'")
                    )
                    escaped_new = (
                        new_str.replace("\t", "    ")
                        .replace("\\", "\\\\")
                        .replace("\n", "\\n")
                        .replace("&", "\\&")
                        .replace("/", "\\/")
                        .replace("'", "'\"'\"'")
                    )
                    cmd = f"sed -i.bak 's/{escaped_old}/{escaped_new}/g' '{escaped_path}'"

                    result = await self._execute_shell_command(cmd)
                    if result["success"]:
                        output = f"Replaced '{old_str[:50]}...' with '{new_str[:50]}...' in {file_path}"
                    else:
                        error = f"Failed to replace text in {file_path}: {result['error']}"
                    logger.debug(f"Action: {action.type} in file {file_path}")

                case "text_editor_insert":
                    # Insert text after specified line
                    file_path = action.path
                    insert_line = action.insert_line
                    new_str = action.new_str

                    # Type guard: path should be str for text editor actions
                    if not isinstance(file_path, str):
                        error = "Invalid path type for text editor insert action.\n"
                        error += f"Failed to insert text in {file_path}: {result['error']}"
                        raise TypeError(error)
                    escaped_path = file_path.replace("'", "'\"'\"'")
                    escaped_content = (
                        new_str.replace("\t", "    ")
                        .replace("\\", "\\\\")
                        .replace("'", "'\"'\"'")
                        .replace("\n", "\\\n")
                    )
                    cmd = f"sed -i.bak '{insert_line}a\\{escaped_content}' '{escaped_path}'"

                    result = await self._execute_shell_command(cmd)
                    if result["success"]:
                        output = f"Inserted text after line {insert_line} in {file_path}"
                    else:
                        error = f"Failed to insert text in {file_path}: {result['error']}"
                    logger.debug(f"Action: {action.type} at line {insert_line} in file {file_path}")

                case _:
                    error = f"Unrecognized action type: {action.type}"
                    logger.warning(error)
        except KeyboardInterrupt:
            error = "User interrupt. Operation aborted."
            logger.error(error)
        except TypeError as e:
            logger.error(f"Error executing action {action.type}: {e}")
        except Exception as e:
            error = f"Unexpected error executing action {action.type}: {str(e)}"
            logger.exception(
                f"Unexpected error during step execution for action: {action.model_dump_json(exclude_none=True)}"
            )

        after_state = await self.get_state()

        if action.type == "screenshot" and step_type == "image":
            output = {"image": after_state.screenshot, "url": after_state.url}

        return EnvStepResult(
            type=step_type,
            output=output,
            error=error,
            current_url=after_state.url,
            screenshot_base64=after_state.screenshot,
        )

    async def _execute_shell_command(self, command: str, new: bool = True) -> dict:
        """Execute a shell command and return the result."""
        try:
            if self.provider == "docker":
                # Execute command in Docker container
                docker_args = [
                    "docker",
                    "exec",
                    self.docker_container_name,
                    "bash",
                    "-c",
                    command,  # The command string is passed as a single argument to bash -c
                ]
                process = await asyncio.to_thread(
                    subprocess.run,
                    docker_args,
                    capture_output=True,
                    text=True,
                    check=False,
                    timeout=120,
                )
            else:
                # Execute command locally
                process = await asyncio.to_thread(
                    subprocess.run,
                    command,
                    shell=True,
                    capture_output=True,
                    text=True,
                    check=False,
                    start_new_session=new,
                    timeout=120,
                )

            if process.returncode == 0:
                return {"success": True, "output": process.stdout, "error": None}
            else:
                return {"success": False, "output": process.stdout, "error": process.stderr}
        except asyncio.TimeoutError:
            return {"success": False, "output": "", "error": "Command timed out after 120 seconds."}
        except Exception as e:
            return {"success": False, "output": "", "error": str(e)}

    async def close(self) -> None:
        logger.debug("Computer environment closed. No specific resources to release for PyAutoGUI.")

    CUA_KEY_TO_PYAUTOGUI_KEY = {
        # Modifiers
        "option": "alt",
        "control": "ctrl",
        "cmd": "command",
        "super": "win",
        "meta": "command" if platform.system() == "Darwin" else "win",
        # Navigation & Editing
        "arrowdown": "down",
        "arrowleft": "left",
        "arrowright": "right",
        "arrowup": "up",
        "caps_lock": "capslock",
        "del": "delete",
        "return": "enter",
        "esc": "escape",
        "pgdn": "pagedown",
        "pgup": "pageup",
        " ": "space",
        # Numpad keys (example, pyautogui uses 'num0', 'add', 'subtract', etc.)
        "numpad0": "num0",
        "numpad_0": "num0",
    }

    @staticmethod
    def parse_key_combination(text: str) -> list[str]:
        if not text:
            return []

        keys_str_list = text.lower().split("+")
        mapped_keys = []
        for k_str in keys_str_list:
            # Use the mapped key if found, otherwise use the string itself (e.g. 'a', '1')
            mapped_keys.append(ComputerEnvironment.CUA_KEY_TO_PYAUTOGUI_KEY.get(k_str.strip(), k_str.strip()))
        return mapped_keys

    def generate_pyautogui_command(self, func_name: str, *args, **kwargs) -> str:
        args_repr = [repr(arg) for arg in args]
        kwargs_repr = [f"{k}={repr(v)}" for k, v in kwargs.items()]
        all_params_repr = ", ".join(args_repr + kwargs_repr)

        # Base script setup
        script_lines = [
            "import os",
            "import pyautogui",
        ]

        if self.provider == "docker":
            script_lines.extend(
                [
                    # Display export for Docker.
                    f"os.environ['DISPLAY']='{self.docker_display}'",
                    # Disable failsafe in Docker to avoid accidental exits
                    "pyautogui.FAILSAFE = False",
                ]
            )

        # Function-specific logic
        if func_name == "screenshot":
            script_lines.extend(
                [
                    "import io",
                    "import base64",
                    "img = pyautogui.screenshot()",
                    "buf = io.BytesIO()",
                    "img.save(buf, format='PNG')",
                    "print(base64.b64encode(buf.getvalue()).decode('utf-8'))",
                ]
            )
        elif func_name == "size":
            script_lines.extend(["size = pyautogui.size()", "print(f'({size.width}, {size.height})')"])
        elif func_name == "position":
            script_lines.extend(["pos = pyautogui.position()", "print(f'({pos.x}, {pos.y})')"])
        else:  # General command structure
            script_lines.extend(
                [f"result = pyautogui.{func_name}({all_params_repr})", "print(result if result is not None else '')"]
            )

        return "; ".join(script_lines)

    async def docker_execute(self, python_command_str: str) -> Optional[str]:
        if not self.docker_container_name or not self.docker_display:
            logger.error("Container name or Docker display not set for Docker execution.")
            return None

        safe_python_cmd = python_command_str.replace('"', '\\"')
        docker_full_cmd = (
            f'docker exec -e DISPLAY={self.docker_display} "{self.docker_container_name}" '
            f'python3 -c "{safe_python_cmd}"'
        )

        try:
            process = await asyncio.to_thread(
                subprocess.run,
                docker_full_cmd,
                shell=True,
                capture_output=True,
                text=True,
                check=False,  # We check returncode manually
            )
            if process.returncode != 0:
                if "FailSafeException" in process.stderr or "FailSafeException" in process.stdout:
                    raise KeyboardInterrupt(process.stderr or process.stdout)
                else:
                    error_msg = (
                        f"Docker command failed:\nCmd: {docker_full_cmd}\n"
                        f"Return Code: {process.returncode}\nStderr: {process.stderr}\nStdout: {process.stdout}"
                    )
                    logger.error(error_msg)
                    raise RuntimeError(f"Docker exec error: {process.stderr or process.stdout}")
            return process.stdout.strip()
        except KeyboardInterrupt:  # Re-raise if caught from above
            raise
        except Exception as e:
            logger.error(f"Unexpected error running command in Docker '{docker_full_cmd}': {e}")
            # Encapsulate as RuntimeError to avoid leaking subprocess errors directly
            raise RuntimeError(f"Unexpected Docker error: {e}") from e



from pathlib import Path
from typing import Dict

app_root_directory = Path(__file__).parent.parent.parent
web_directory = app_root_directory / "khoj/interface/web/"
next_js_directory = app_root_directory / "khoj/interface/built/"
pypi_static_directory = app_root_directory / "khoj/interface/compiled/"
assetlinks_file_path = web_directory / ".well-known/assetlinks.json"
empty_escape_sequences = "\n|\r|\t| "
app_env_filepath = "~/.khoj/env"
telemetry_server = "https://khoj.beta.haletic.com/v1/telemetry"
content_directory = "~/.khoj/content/"
default_openai_chat_models = ["gpt-4o-mini", "gpt-4.1", "o3", "o4-mini"]
default_gemini_chat_models = ["gemini-2.0-flash", "gemini-2.5-flash-preview-05-20", "gemini-2.5-pro-preview-06-05"]
default_anthropic_chat_models = ["claude-sonnet-4-0", "claude-3-5-haiku-latest"]

empty_config = {
    "search-type": {
        "image": {"encoder": "sentence-transformers/clip-ViT-B-32", "model_directory": "~/.khoj/search/image/"},
    },
}

# default app config to use
default_config = {
    "search-type": {
        "image": {"encoder": "sentence-transformers/clip-ViT-B-32", "model_directory": "~/.khoj/search/image/"},
    },
}

model_to_cost: Dict[str, Dict[str, float]] = {
    # OpenAI Pricing: https://openai.com/api/pricing/
    "gpt-4o": {"input": 2.50, "output": 10.00},
    "gpt-4o-mini": {"input": 0.15, "output": 0.60},
    "gpt-4.1": {"input": 2.00, "output": 8.00},
    "gpt-4.1-mini": {"input": 0.40, "output": 1.60},
    "gpt-4.1-nano": {"input": 0.10, "output": 0.40},
    "o1-mini": {"input": 3.0, "output": 12.0},
    "o1": {"input": 15.0, "output": 60.00},
    "o3-mini": {"input": 1.10, "output": 4.40},
    "o3": {"input": 2.0, "output": 8.00},
    "o3-pro": {"input": 20.0, "output": 80.00},
    "o4-mini": {"input": 1.10, "output": 4.40},
    "gpt-5-2025-08-07": {"input": 1.25, "output": 10.00, "cache_read": 0.125},
    "gpt-5-mini-2025-08-07": {"input": 0.25, "output": 2.00, "cache_read": 0.025},
    "gpt-5-nano-2025-08-07": {"input": 0.05, "output": 0.40, "cache_read": 0.005},
    # Gemini Pricing: https://ai.google.dev/pricing
    "gemini-1.5-flash": {"input": 0.075, "output": 0.30},
    "gemini-1.5-flash-002": {"input": 0.075, "output": 0.30},
    "gemini-1.5-pro": {"input": 1.25, "output": 5.00},
    "gemini-1.5-pro-002": {"input": 1.25, "output": 5.00},
    "gemini-2.0-flash": {"input": 0.10, "output": 0.40},
    "gemini-2.0-flash-lite": {"input": 0.0075, "output": 0.30},
    "gemini-2.5-flash": {"input": 0.30, "output": 2.50},
    "gemini-2.5-pro": {"input": 1.25, "output": 10.0},
    # Anthropic Pricing: https://www.anthropic.com/pricing#anthropic-api
    "claude-3-5-haiku-20241022": {"input": 1.0, "output": 5.0, "cache_read": 0.08, "cache_write": 1.0},
    "claude-3-5-haiku@20241022": {"input": 1.0, "output": 5.0, "cache_read": 0.08, "cache_write": 1.0},
    "claude-3-5-sonnet-20241022": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-3-5-sonnet-latest": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-3-7-sonnet-20250219": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-3-7-sonnet@20250219": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-3-7-sonnet-latest": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-sonnet-4-0": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-sonnet-4-20250514": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-sonnet-4@20250514": {"input": 3.0, "output": 15.0, "cache_read": 0.3, "cache_write": 3.75},
    "claude-opus-4-0": {"input": 15.0, "output": 75.0, "cache_read": 1.50, "cache_write": 18.75},
    "claude-opus-4-20250514": {"input": 15.0, "output": 75.0, "cache_read": 1.50, "cache_write": 18.75},
    "claude-opus-4@20250514": {"input": 15.0, "output": 75.0, "cache_read": 1.50, "cache_write": 18.75},
    # Grok pricing: https://docs.x.ai/docs/models
    "grok-3": {"input": 3.0, "output": 15.0},
    "grok-3-latest": {"input": 3.0, "output": 15.0},
    "grok-3-mini": {"input": 0.30, "output": 0.50},
    "grok-3-mini-latest": {"input": 0.30, "output": 0.50},
    # Groq pricing
    "moonshotai/kimi-k2-instruct": {"input": 1.00, "output": 3.00},
    "openai/gpt-oss-120b": {"input": 0.15, "output": 0.75},
}



import yaml

# Do not emit tags when dumping to YAML
yaml.emitter.Emitter.process_tag = lambda self, *args, **kwargs: None  # type: ignore[assignment]


def yaml_dump(data):
    return yaml.dump(data, allow_unicode=True, sort_keys=False, default_flow_style=False)



import json
import logging
import threading
from typing import Optional

import cron_descriptor
import pytz
from apscheduler.job import Job
from apscheduler.triggers.cron import CronTrigger
from fastapi import APIRouter, Request
from fastapi.responses import Response
from starlette.authentication import requires

from khoj.database.adapters import AutomationAdapters, ConversationAdapters
from khoj.database.models import KhojUser
from khoj.processor.conversation.utils import clean_json
from khoj.routers.helpers import schedule_automation, schedule_query
from khoj.utils.helpers import is_none_or_empty

# Initialize Router
api_automation = APIRouter()
logger = logging.getLogger(__name__)


@api_automation.get("", response_class=Response)
@requires(["authenticated"])
def get_automations(request: Request) -> Response:
    user: KhojUser = request.user.object

    # Collate all automations created by user that are still active
    automations_info = [automation_info for automation_info in AutomationAdapters.get_automations_metadata(user)]

    # Return tasks information as a JSON response
    return Response(content=json.dumps(automations_info), media_type="application/json", status_code=200)


@api_automation.delete("", response_class=Response)
@requires(["authenticated"])
def delete_automation(request: Request, automation_id: str) -> Response:
    user: KhojUser = request.user.object

    try:
        automation_info = AutomationAdapters.delete_automation(user, automation_id)
    except ValueError:
        return Response(status_code=204)

    # Return deleted automation information as a JSON response
    return Response(content=json.dumps(automation_info), media_type="application/json", status_code=200)


@api_automation.post("", response_class=Response)
@requires(["authenticated"])
def post_automation(
    request: Request,
    q: str,
    crontime: str,
    subject: Optional[str] = None,
    city: Optional[str] = None,
    region: Optional[str] = None,
    country: Optional[str] = None,
    timezone: Optional[str] = None,
) -> Response:
    user: KhojUser = request.user.object

    # Perform validation checks
    if is_none_or_empty(q) or is_none_or_empty(crontime):
        return Response(content="A query and crontime is required", status_code=400)
    if not cron_descriptor.get_description(crontime):
        return Response(content="Invalid crontime", status_code=400)

    # Infer subject, query to run
    _, query_to_run, generated_subject = schedule_query(q, chat_history=[], user=user)
    subject = subject or generated_subject

    # Normalize query parameters
    # Add /automated_task prefix to query if not present
    query_to_run = query_to_run.strip()
    if not query_to_run.startswith("/automated_task"):
        query_to_run = f"/automated_task {query_to_run}"

    # Normalize crontime for AP Scheduler CronTrigger
    crontime = crontime.strip()
    if len(crontime.split(" ")) > 5:
        # Truncate crontime to 5 fields
        crontime = " ".join(crontime.split(" ")[:5])

    # Convert crontime to standard unix crontime
    crontime = crontime.replace("?", "*")

    # Disallow minute level automation recurrence
    minute_value = crontime.split(" ")[0]
    if not minute_value.isdigit():
        return Response(
            content="Minute level recurrence is unsupported. Please create a less frequent schedule.",
            status_code=400,
        )

    # Create new Conversation Session associated with this new task
    title = f"Automation: {subject}"
    conversation = ConversationAdapters.create_conversation_session(user, request.user.client_app, title=title)

    # Schedule automation with query_to_run, timezone, subject directly provided by user
    try:
        # Use the query to run as the scheduling request if the scheduling request is unset
        calling_url = str(request.url.replace(query=f"{request.url.query}"))
        automation = schedule_automation(
            query_to_run, subject, crontime, timezone, q, user, calling_url, str(conversation.id)
        )
    except Exception as e:
        logger.error(f"Error creating automation {q} for {user.email}: {e}", exc_info=True)
        return Response(
            content="Unable to create automation. Ensure the automation doesn't already exist.",
            media_type="text/plain",
            status_code=500,
        )

    # Collate info about the created user automation
    automation_info = AutomationAdapters.get_automation_metadata(user, automation)

    # Return information about the created automation as a JSON response
    return Response(content=json.dumps(automation_info), media_type="application/json", status_code=200)


@api_automation.post("/trigger", response_class=Response)
@requires(["authenticated"])
def trigger_manual_job(
    request: Request,
    automation_id: str,
):
    user: KhojUser = request.user.object

    # Check, get automation to edit
    try:
        automation: Job = AutomationAdapters.get_automation(user, automation_id)
    except ValueError as e:
        logger.error(f"Error triggering automation {automation_id} for {user.email}: {e}", exc_info=True)
        return Response(content="Invalid automation", status_code=403)

    # Trigger the job without waiting for the result.
    scheduled_chat_func = automation.func

    # Run the function in a separate thread
    thread = threading.Thread(target=scheduled_chat_func, args=automation.args, kwargs=automation.kwargs)
    thread.start()

    return Response(content="Automation triggered", status_code=200)


@api_automation.put("", response_class=Response)
@requires(["authenticated"])
def edit_job(
    request: Request,
    automation_id: str,
    q: Optional[str],
    subject: Optional[str],
    crontime: Optional[str],
    city: Optional[str] = None,
    region: Optional[str] = None,
    country: Optional[str] = None,
    timezone: Optional[str] = None,
) -> Response:
    user: KhojUser = request.user.object

    # Perform validation checks
    if is_none_or_empty(q) or is_none_or_empty(subject) or is_none_or_empty(crontime):
        return Response(content="A query, subject and crontime is required", status_code=400)
    if not cron_descriptor.get_description(crontime):
        return Response(content="Invalid crontime", status_code=400)

    # Check, get automation to edit
    try:
        automation: Job = AutomationAdapters.get_automation(user, automation_id)
    except ValueError as e:
        logger.error(f"Error editing automation {automation_id} for {user.email}: {e}", exc_info=True)
        return Response(content="Invalid automation", status_code=403)

    # Infer subject, query to run
    _, query_to_run, _ = schedule_query(q, chat_history=[], user=user)
    subject = subject

    # Normalize query parameters
    # Add /automated_task prefix to query if not present
    query_to_run = query_to_run.strip()
    if not query_to_run.startswith("/automated_task"):
        query_to_run = f"/automated_task {query_to_run}"
    # Normalize crontime for AP Scheduler CronTrigger
    crontime = crontime.strip()
    if len(crontime.split(" ")) > 5:
        # Truncate crontime to 5 fields
        crontime = " ".join(crontime.split(" ")[:5])
    # Convert crontime to standard unix crontime
    crontime = crontime.replace("?", "*")

    # Disallow minute level automation recurrence
    minute_value = crontime.split(" ")[0]
    if not minute_value.isdigit():
        return Response(
            content="Recurrence of every X minutes is unsupported. Please create a less frequent schedule.",
            status_code=400,
        )

    # Construct updated automation metadata
    automation_metadata: dict[str, str] = json.loads(clean_json(automation.name))
    automation_metadata["scheduling_request"] = q
    automation_metadata["query_to_run"] = query_to_run
    automation_metadata["subject"] = subject.strip()
    automation_metadata["crontime"] = crontime
    conversation_id = automation_metadata.get("conversation_id")

    if not conversation_id:
        title = f"Automation: {subject}"

        # Create new Conversation Session associated with this new task
        conversation = ConversationAdapters.create_conversation_session(user, request.user.client_app, title=title)

        conversation_id = str(conversation.id)
        automation_metadata["conversation_id"] = conversation_id

    # Modify automation with updated query, subject
    automation.modify(
        name=json.dumps(automation_metadata),
        kwargs={
            "query_to_run": query_to_run,
            "subject": subject,
            "scheduling_request": q,
            "user": user,
            "calling_url": str(request.url),
            "conversation_id": conversation_id,
        },
    )

    # Reschedule automation if crontime updated
    user_timezone = pytz.timezone(timezone)
    trigger = CronTrigger.from_crontab(crontime, user_timezone)
    if automation.trigger != trigger:
        automation.reschedule(trigger=trigger)

    # Collate info about the updated user automation
    automation = AutomationAdapters.get_automation(user, automation.id)
    automation_info = AutomationAdapters.get_automation_metadata(user, automation)

    # Return modified automation information as a JSON response
    return Response(content=json.dumps(automation_info), media_type="application/json", status_code=200)



# Generated by Django 4.2.5 on 2023-10-13 02:39

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0004_content_types_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="embeddings",
            name="corpus_id",
            field=models.UUIDField(default=uuid.uuid4, editable=False),
        ),
    ]



import asyncio
import logging
import os
from copy import deepcopy
from datetime import datetime
from typing import Callable, Dict, List, Optional

import yaml

from khoj.database.adapters import AgentAdapters, EntryAdapters
from khoj.database.models import Agent, ChatMessageModel, KhojUser
from khoj.processor.conversation import prompts
from khoj.processor.conversation.utils import (
    OperatorRun,
    ResearchIteration,
    ToolCall,
    construct_iteration_history,
    construct_tool_chat_history,
    load_complex_json,
)
from khoj.processor.operator import operate_environment
from khoj.processor.tools.online_search import read_webpages_content, search_online
from khoj.processor.tools.run_code import run_code
from khoj.routers.helpers import (
    ChatEvent,
    get_message_from_queue,
    grep_files,
    list_files,
    search_documents,
    send_message_to_model_wrapper,
    view_file_content,
)
from khoj.utils.helpers import (
    ConversationCommand,
    ToolDefinition,
    dict_to_tuple,
    is_none_or_empty,
    is_operator_enabled,
    timer,
    tools_for_research_llm,
    truncate_code_context,
)
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)


async def apick_next_tool(
    query: str,
    conversation_history: List[ChatMessageModel],
    user: KhojUser = None,
    location: LocationData = None,
    user_name: str = None,
    agent: Agent = None,
    previous_iterations: List[ResearchIteration] = [],
    max_iterations: int = 5,
    query_images: List[str] = [],
    query_files: str = None,
    max_document_searches: int = 7,
    max_online_searches: int = 3,
    max_webpages_to_read: int = 3,
    send_status_func: Optional[Callable] = None,
    tracer: dict = {},
):
    """Given a query, determine which of the available tools the agent should use in order to answer appropriately."""

    # Continue with previous iteration if a multi-step tool use is in progress
    if (
        previous_iterations
        and previous_iterations[-1].query
        and isinstance(previous_iterations[-1].query, ToolCall)
        and previous_iterations[-1].query.name == ConversationCommand.Operator
        and not previous_iterations[-1].summarizedResult
    ):
        previous_iteration = previous_iterations[-1]
        yield ResearchIteration(
            query=ToolCall(name=previous_iteration.query.name, args={"query": query}, id=previous_iteration.query.id),  # type: ignore
            context=previous_iteration.context,
            onlineContext=previous_iteration.onlineContext,
            codeContext=previous_iteration.codeContext,
            operatorContext=previous_iteration.operatorContext,
            warning=previous_iteration.warning,
        )
        return

    # Construct tool options for the agent to choose from
    tools = []
    tool_options_str = ""
    agent_input_tools = agent.input_tools if agent else []
    agent_tools = []

    # Map agent user facing tools to research tools to include in agents toolbox
    document_research_tools = [
        ConversationCommand.SemanticSearchFiles,
        ConversationCommand.RegexSearchFiles,
        ConversationCommand.ViewFile,
        ConversationCommand.ListFiles,
    ]
    input_tools_to_research_tools = {
        ConversationCommand.Notes.value: [tool.value for tool in document_research_tools],
        ConversationCommand.Webpage.value: [ConversationCommand.ReadWebpage.value],
        ConversationCommand.Online.value: [ConversationCommand.SearchWeb.value],
        ConversationCommand.Code.value: [ConversationCommand.PythonCoder.value],
        ConversationCommand.Operator.value: [ConversationCommand.OperateComputer.value],
    }
    for input_tool, research_tools in input_tools_to_research_tools.items():
        if input_tool in agent_input_tools:
            agent_tools += research_tools

    user_has_entries = await EntryAdapters.auser_has_entries(user)
    for tool, tool_data in tools_for_research_llm.items():
        # Skip showing operator tool as an option if not enabled
        if tool == ConversationCommand.OperateComputer and not is_operator_enabled():
            continue
        # Skip showing document related tools if user has no documents
        if tool in document_research_tools and not user_has_entries:
            continue
        if tool == ConversationCommand.SemanticSearchFiles:
            description = tool_data.description.format(max_search_queries=max_document_searches)
        elif tool == ConversationCommand.ReadWebpage:
            description = tool_data.description.format(max_webpages_to_read=max_webpages_to_read)
        elif tool == ConversationCommand.SearchWeb:
            description = tool_data.description.format(max_search_queries=max_online_searches)
        else:
            description = tool_data.description
        # Add tool if agent does not have any tools defined or the tool is supported by the agent.
        if len(agent_tools) == 0 or tool.value in agent_tools:
            tool_options_str += f'- "{tool.value}": "{description}"\n'
            tools.append(
                ToolDefinition(
                    name=tool.value,
                    description=description,
                    schema=tool_data.schema,
                )
            )

    today = datetime.today()
    location_data = f"{location}" if location else "Unknown"
    agent_chat_model = AgentAdapters.get_agent_chat_model(agent, user) if agent else None
    personality_context = (
        prompts.personality_context.format(personality=agent.personality) if agent and agent.personality else ""
    )

    function_planning_prompt = prompts.plan_function_execution.format(
        tools=tool_options_str,
        personality_context=personality_context,
        current_date=today.strftime("%Y-%m-%d"),
        day_of_week=today.strftime("%A"),
        username=user_name or "Unknown",
        location=location_data,
        max_iterations=max_iterations,
    )

    # Construct chat history with user and iteration history with researcher agent for context
    iteration_chat_history = construct_iteration_history(previous_iterations, query, query_images, query_files)
    chat_and_research_history = conversation_history + iteration_chat_history

    try:
        with timer("Chat actor: Infer information sources to refer", logger):
            response = await send_message_to_model_wrapper(
                query="",
                system_message=function_planning_prompt,
                chat_history=chat_and_research_history,
                tools=tools,
                deepthought=True,
                user=user,
                query_images=query_images,
                query_files=query_files,
                agent_chat_model=agent_chat_model,
                tracer=tracer,
            )
    except Exception as e:
        logger.error(f"Failed to infer information sources to refer: {e}", exc_info=True)
        yield ResearchIteration(
            query=None,
            warning="Failed to infer information sources to refer. Skipping iteration. Try again.",
        )
        return

    try:
        # Try parse the response as function call response to infer next tool to use.
        # TODO: Handle multiple tool calls.
        response_text = response.text
        parsed_response = [ToolCall(**item) for item in load_complex_json(response_text)][0]
    except Exception:
        # Otherwise assume the model has decided to end the research run and respond to the user.
        parsed_response = ToolCall(name=ConversationCommand.Text, args={"response": response_text}, id=None)

    # If we have a valid response, extract the tool and query.
    warning = None
    logger.info(f"Response for determining relevant tools: {parsed_response.name}({parsed_response.args})")

    # Detect selection of previously used query, tool combination.
    previous_tool_query_combinations = {
        (i.query.name, dict_to_tuple(i.query.args))
        for i in previous_iterations
        if i.warning is None and isinstance(i.query, ToolCall)
    }
    if (parsed_response.name, dict_to_tuple(parsed_response.args)) in previous_tool_query_combinations:
        warning = f"Repeated tool, query combination detected. You've already called {parsed_response.name} with args: {parsed_response.args}. Try something different."
    # Only send client status updates if we'll execute this iteration and model has thoughts to share.
    elif send_status_func and not is_none_or_empty(response.thought):
        async for event in send_status_func(response.thought):
            yield {ChatEvent.STATUS: event}

    yield ResearchIteration(query=parsed_response, warning=warning, raw_response=response.raw_content)


async def research(
    user: KhojUser,
    query: str,
    conversation_id: str,
    conversation_history: List[ChatMessageModel],
    previous_iterations: List[ResearchIteration],
    query_images: List[str],
    agent: Agent = None,
    send_status_func: Optional[Callable] = None,
    user_name: str = None,
    location: LocationData = None,
    file_filters: List[str] = [],
    tracer: dict = {},
    query_files: str = None,
    cancellation_event: Optional[asyncio.Event] = None,
    interrupt_queue: Optional[asyncio.Queue] = None,
    abort_message: str = ChatEvent.END_EVENT.value,
):
    max_document_searches = 7
    max_online_searches = 3
    max_webpages_to_read = 1
    current_iteration = 0
    MAX_ITERATIONS = int(os.getenv("KHOJ_RESEARCH_ITERATIONS", 5))

    # Incorporate previous partial research into current research chat history
    research_conversation_history = [chat for chat in deepcopy(conversation_history) if chat.message]
    if current_iteration := len(previous_iterations) > 0:
        logger.info(f"Continuing research with the previous {len(previous_iterations)} iteration results.")
        previous_iterations_history = construct_iteration_history(previous_iterations)
        research_conversation_history += previous_iterations_history

    while current_iteration < MAX_ITERATIONS:
        # Check for cancellation at the start of each iteration
        if cancellation_event and cancellation_event.is_set():
            logger.debug(f"Research cancelled. User {user} disconnected client.")
            break

        # Update the query for the current research iteration
        if interrupt_query := get_message_from_queue(interrupt_queue):
            if interrupt_query == abort_message:
                cancellation_event.set()
                logger.debug(f"Research cancelled by user {user} via interrupt queue.")
                break
            # Add the interrupt query as a new user message to the research conversation history
            logger.info(
                f"Continuing research for user {user} with the previous {len(previous_iterations)} iterations and new instruction: {interrupt_query}"
            )
            previous_iterations_history = construct_iteration_history(
                previous_iterations, query, query_images, query_files
            )
            research_conversation_history += previous_iterations_history
            query = interrupt_query
            previous_iterations = []

            async for result in send_status_func(f"**Incorporate New Instruction**: {interrupt_query}"):
                yield result

        online_results: Dict = dict()
        code_results: Dict = dict()
        document_results: List[Dict[str, str]] = []
        operator_results: OperatorRun = None
        this_iteration = ResearchIteration(query=query)

        async for result in apick_next_tool(
            query,
            research_conversation_history,
            user,
            location,
            user_name,
            agent,
            previous_iterations,
            MAX_ITERATIONS,
            query_images=query_images,
            query_files=query_files,
            max_document_searches=max_document_searches,
            max_online_searches=max_online_searches,
            max_webpages_to_read=max_webpages_to_read,
            send_status_func=send_status_func,
            tracer=tracer,
        ):
            if isinstance(result, dict) and ChatEvent.STATUS in result:
                yield result[ChatEvent.STATUS]
            elif isinstance(result, ResearchIteration):
                this_iteration = result
                yield this_iteration

        # Skip running iteration if warning present in iteration
        if this_iteration.warning:
            logger.warning(f"Research mode: {this_iteration.warning}.")

        # Terminate research if selected text tool or query, tool not set for next iteration
        elif (
            not this_iteration.query
            or isinstance(this_iteration.query, str)
            or this_iteration.query.name == ConversationCommand.Text
        ):
            current_iteration = MAX_ITERATIONS

        elif this_iteration.query.name == ConversationCommand.SemanticSearchFiles:
            this_iteration.context = []
            document_results = []
            previous_inferred_queries = {
                c["query"] for iteration in previous_iterations if iteration.context for c in iteration.context
            }
            async for result in search_documents(
                **this_iteration.query.args,
                n=max_document_searches,
                d=None,
                user=user,
                chat_history=construct_tool_chat_history(previous_iterations, ConversationCommand.SemanticSearchFiles),
                conversation_id=conversation_id,
                conversation_commands=[ConversationCommand.Default],
                location_data=location,
                send_status_func=send_status_func,
                query_images=query_images,
                previous_inferred_queries=previous_inferred_queries,
                agent=agent,
                tracer=tracer,
                query_files=query_files,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                elif isinstance(result, tuple):
                    document_results = result[0]
                    this_iteration.context += document_results

            if not is_none_or_empty(document_results):
                try:
                    distinct_files = {d["file"] for d in document_results}
                    distinct_headings = set([d["compiled"].split("\n")[0] for d in document_results if "compiled" in d])
                    # Strip only leading # from headings
                    headings_str = "\n- " + "\n- ".join(distinct_headings).replace("#", "")
                    async for result in send_status_func(
                        f"**Found {len(distinct_headings)} Notes Across {len(distinct_files)} Files**: {headings_str}"
                    ):
                        yield result
                except Exception as e:
                    this_iteration.warning = f"Error extracting document references: {e}"
                    logger.error(this_iteration.warning, exc_info=True)
            else:
                this_iteration.warning = "No matching document references found"

        elif this_iteration.query.name == ConversationCommand.SearchWeb:
            previous_subqueries = {
                subquery
                for iteration in previous_iterations
                if iteration.onlineContext
                for subquery in iteration.onlineContext.keys()
            }
            try:
                async for result in search_online(
                    **this_iteration.query.args,
                    conversation_history=construct_tool_chat_history(
                        previous_iterations, ConversationCommand.SearchWeb
                    ),
                    location=location,
                    user=user,
                    send_status_func=send_status_func,
                    custom_filters=[],
                    max_online_searches=max_online_searches,
                    max_webpages_to_read=0,
                    query_images=query_images,
                    previous_subqueries=previous_subqueries,
                    agent=agent,
                    tracer=tracer,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    elif is_none_or_empty(result):
                        this_iteration.warning = "Detected previously run online search queries. Skipping iteration. Try something different."
                    else:
                        online_results: Dict[str, Dict] = result  # type: ignore
                        this_iteration.onlineContext = online_results
            except Exception as e:
                this_iteration.warning = f"Error searching online: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.ReadWebpage:
            try:
                async for result in read_webpages_content(
                    **this_iteration.query.args,
                    user=user,
                    send_status_func=send_status_func,
                    # max_webpages_to_read=max_webpages_to_read,
                    agent=agent,
                    tracer=tracer,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    else:
                        direct_web_pages: Dict[str, Dict] = result  # type: ignore

                        webpages = []
                        for web_query in direct_web_pages:
                            if online_results.get(web_query):
                                online_results[web_query]["webpages"] = direct_web_pages[web_query]["webpages"]
                            else:
                                online_results[web_query] = {"webpages": direct_web_pages[web_query]["webpages"]}

                            for webpage in direct_web_pages[web_query]["webpages"]:
                                webpages.append(webpage["link"])
                        this_iteration.onlineContext = online_results
            except Exception as e:
                this_iteration.warning = f"Error reading webpages: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.PythonCoder:
            try:
                async for result in run_code(
                    **this_iteration.query.args,
                    conversation_history=construct_tool_chat_history(
                        previous_iterations, ConversationCommand.PythonCoder
                    ),
                    context="",
                    location_data=location,
                    user=user,
                    send_status_func=send_status_func,
                    query_images=query_images,
                    agent=agent,
                    query_files=query_files,
                    tracer=tracer,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    else:
                        code_results: Dict[str, Dict] = result  # type: ignore
                        this_iteration.codeContext = code_results
                async for result in send_status_func(f"**Ran code snippets**: {len(this_iteration.codeContext)}"):
                    yield result
            except (ValueError, TypeError) as e:
                this_iteration.warning = f"Error running code: {e}"
                logger.warning(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.OperateComputer:
            try:
                async for result in operate_environment(
                    **this_iteration.query.args,
                    user=user,
                    conversation_log=construct_tool_chat_history(previous_iterations, ConversationCommand.Operator),
                    location_data=location,
                    previous_trajectory=previous_iterations[-1].operatorContext if previous_iterations else None,
                    send_status_func=send_status_func,
                    query_images=query_images,
                    agent=agent,
                    query_files=query_files,
                    cancellation_event=cancellation_event,
                    interrupt_queue=interrupt_queue,
                    tracer=tracer,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    elif isinstance(result, OperatorRun):
                        operator_results = result
                        this_iteration.operatorContext = operator_results
                        # Add webpages visited while operating browser to references
                        if result.webpages:
                            if not online_results.get(this_iteration.query):
                                online_results[this_iteration.query] = {"webpages": result.webpages}
                            elif not online_results[this_iteration.query].get("webpages"):
                                online_results[this_iteration.query]["webpages"] = result.webpages
                            else:
                                online_results[this_iteration.query]["webpages"] += result.webpages
                            this_iteration.onlineContext = online_results
            except Exception as e:
                this_iteration.warning = f"Error operating browser: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.ViewFile:
            try:
                async for result in view_file_content(
                    **this_iteration.query.args,
                    user=user,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    else:
                        if this_iteration.context is None:
                            this_iteration.context = []
                        document_results: List[Dict[str, str]] = result  # type: ignore
                        this_iteration.context += document_results
                async for result in send_status_func(f"**Viewed file**: {this_iteration.query.args['path']}"):
                    yield result
            except Exception as e:
                this_iteration.warning = f"Error viewing file: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.ListFiles:
            try:
                async for result in list_files(
                    **this_iteration.query.args,
                    user=user,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    else:
                        if this_iteration.context is None:
                            this_iteration.context = []
                        document_results: List[Dict[str, str]] = [result]  # type: ignore
                        this_iteration.context += document_results
                async for result in send_status_func(result["query"]):
                    yield result
            except Exception as e:
                this_iteration.warning = f"Error listing files: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        elif this_iteration.query.name == ConversationCommand.RegexSearchFiles:
            try:
                async for result in grep_files(
                    **this_iteration.query.args,
                    user=user,
                ):
                    if isinstance(result, dict) and ChatEvent.STATUS in result:
                        yield result[ChatEvent.STATUS]
                    else:
                        if this_iteration.context is None:
                            this_iteration.context = []
                        document_results: List[Dict[str, str]] = [result]  # type: ignore
                        this_iteration.context += document_results
                async for result in send_status_func(result["query"]):
                    yield result
            except Exception as e:
                this_iteration.warning = f"Error searching with regex: {e}"
                logger.error(this_iteration.warning, exc_info=True)

        else:
            # No valid tools. This is our exit condition.
            current_iteration = MAX_ITERATIONS

        current_iteration += 1

        if document_results or online_results or code_results or operator_results or this_iteration.warning:
            results_data = f"\n<iteration_{current_iteration}_results>"
            if document_results:
                results_data += f"\n<document_references>\n{yaml.dump(document_results, allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</document_references>"
            if online_results:
                results_data += f"\n<online_results>\n{yaml.dump(online_results, allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</online_results>"
            if code_results:
                results_data += f"\n<code_results>\n{yaml.dump(truncate_code_context(code_results), allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</code_results>"
            if operator_results:
                results_data += (
                    f"\n<browser_operator_results>\n{operator_results.response}\n</browser_operator_results>"
                )
            if this_iteration.warning:
                results_data += f"\n<warning>\n{this_iteration.warning}\n</warning>"
            results_data += f"\n</results>\n</iteration_{current_iteration}_results>"

            # intermediate_result = await extract_relevant_info(this_iteration.query, results_data, agent)
            this_iteration.summarizedResult = results_data

        this_iteration.summarizedResult = this_iteration.summarizedResult or "Failed to get results."
        previous_iterations.append(this_iteration)
        yield this_iteration



# Generated by Django 5.1.10 on 2025-06-28 06:50

from django.db import migrations, models


def initialize_friendly_names(apps, schema_editor):
    """Initialize friendly_name fields with values from the name/model_name fields"""
    ChatModel = apps.get_model("database", "ChatModel")
    SpeechToTextModelOptions = apps.get_model("database", "SpeechToTextModelOptions")
    TextToImageModelConfig = apps.get_model("database", "TextToImageModelConfig")

    # Initialize ChatModel friendly_name with name field
    ChatModel.objects.update(friendly_name=models.F("name"))

    # Initialize SpeechToTextModelOptions friendly_name with model_name field
    SpeechToTextModelOptions.objects.update(friendly_name=models.F("model_name"))

    # Initialize TextToImageModelConfig friendly_name with model_name field
    TextToImageModelConfig.objects.update(friendly_name=models.F("model_name"))


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0090_alter_khojuser_uuid"),
    ]

    operations = [
        migrations.AddField(
            model_name="chatmodel",
            name="friendly_name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name="speechtotextmodeloptions",
            name="friendly_name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name="texttoimagemodelconfig",
            name="friendly_name",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.RunPython(initialize_friendly_names, reverse_code=migrations.RunPython.noop),
    ]



import logging
import re
from typing import Dict, List, Tuple

import urllib3
from bs4 import BeautifulSoup

from khoj.database.models import Entry as DbEntry
from khoj.database.models import KhojUser
from khoj.processor.content.text_to_entries import TextToEntries
from khoj.utils.helpers import timer
from khoj.utils.rawconfig import Entry

logger = logging.getLogger(__name__)


class PlaintextToEntries(TextToEntries):
    def __init__(self):
        super().__init__()

    # Define Functions
    def process(self, files: dict[str, str], user: KhojUser, regenerate: bool = False) -> Tuple[int, int]:
        deletion_file_names = set([file for file in files if files[file] == ""])
        files_to_process = set(files) - deletion_file_names
        files = {file: files[file] for file in files_to_process}

        # Extract Entries from specified plaintext files
        with timer("Extract entries from specified Plaintext files", logger):
            file_to_text_map, current_entries = PlaintextToEntries.extract_plaintext_entries(files)

        # Split entries by max tokens supported by model
        with timer("Split entries by max token size supported by model", logger):
            current_entries = self.split_entries_by_max_tokens(current_entries, max_tokens=256, raw_is_compiled=True)

        # Identify, mark and merge any new entries with previous entries
        with timer("Identify new or updated entries", logger):
            num_new_embeddings, num_deleted_embeddings = self.update_embeddings(
                user,
                current_entries,
                DbEntry.EntryType.PLAINTEXT,
                DbEntry.EntrySource.COMPUTER,
                key="compiled",
                logger=logger,
                deletion_filenames=deletion_file_names,
                regenerate=regenerate,
                file_to_text_map=file_to_text_map,
            )

        return num_new_embeddings, num_deleted_embeddings

    @staticmethod
    def extract_html_content(markup_content: str, markup_type: str):
        "Extract content from HTML"
        if markup_type == "xml":
            soup = BeautifulSoup(markup_content, "xml")
        else:
            soup = BeautifulSoup(markup_content, "html.parser")
        return soup.get_text(strip=True, separator="\n")

    @staticmethod
    def extract_plaintext_entries(text_files: Dict[str, str]) -> Tuple[Dict, List[Entry]]:
        entries: List[str] = []
        entry_to_file_map: List[Tuple[str, str]] = []
        file_to_text_map = dict()
        for text_file in text_files:
            try:
                text_content = text_files[text_file]
                entries, entry_to_file_map = PlaintextToEntries.process_single_plaintext_file(
                    text_content, text_file, entries, entry_to_file_map
                )
                file_to_text_map[text_file] = text_content
            except Exception as e:
                logger.warning(f"Unable to read file: {text_file} as plaintext. Skipping file.")
                logger.warning(e, exc_info=True)

        # Extract Entries from specified plaintext files
        return file_to_text_map, PlaintextToEntries.convert_text_files_to_entries(entries, dict(entry_to_file_map))

    @staticmethod
    def process_single_plaintext_file(
        text_content: str,
        text_file: str,
        entries: List[str],
        entry_to_file_map: List[Tuple[str, str]],
    ) -> Tuple[List[str], List[Tuple[str, str]]]:
        if text_file.endswith(("html", "htm", "xml")):
            text_content = PlaintextToEntries.extract_html_content(text_content, text_file.split(".")[-1])
        entry_to_file_map += [(text_content, text_file)]
        entries.extend([text_content])
        return entries, entry_to_file_map

    @staticmethod
    def convert_text_files_to_entries(parsed_entries: List[str], entry_to_file_map: dict[str, str]) -> List[Entry]:
        "Convert each plaintext file into an entry"
        entries: List[Entry] = []
        for parsed_entry in parsed_entries:
            raw_filename = entry_to_file_map[parsed_entry]
            # Check if raw_filename is a URL. If so, save it as is. If not, convert it to a Path.
            if isinstance(raw_filename, str) and re.search(r"^https?://", raw_filename):
                # Escape the URL to avoid issues with special characters
                entry_filename = urllib3.util.parse_url(raw_filename).url
            else:
                entry_filename = raw_filename

            # Append base filename to compiled entry for context to model
            entries.append(
                Entry(
                    raw=parsed_entry,
                    file=f"{entry_filename}",
                    compiled=f"{entry_filename}\n{parsed_entry}",
                    heading=entry_filename,
                )
            )

        logger.debug(f"Converted {len(parsed_entries)} plaintext files to entries")
        return entries



# System Packages
import json
import os

from fastapi import APIRouter, Request
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse
from fastapi.templating import Jinja2Templates
from starlette.authentication import requires

from khoj.database.adapters import get_user_github_config
from khoj.routers.helpers import get_next_url, get_user_config
from khoj.utils import constants
from khoj.utils.rawconfig import GithubContentConfig, GithubRepoConfig

# Initialize Router
web_client = APIRouter()
templates = Jinja2Templates([constants.web_directory, constants.next_js_directory, constants.pypi_static_directory])


# Create Routes
@web_client.get("/", response_class=FileResponse)
def index(request: Request):
    return templates.TemplateResponse("index.html", context={"request": request})


@web_client.post("/", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def index_post(request: Request):
    return templates.TemplateResponse("index.html", context={"request": request})


@web_client.get("/search", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def search_page(request: Request):
    return templates.TemplateResponse(
        "search/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/chat", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def chat_page(request: Request):
    return templates.TemplateResponse(
        "chat/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/login", response_class=FileResponse)
def login_page(request: Request):
    next_url = get_next_url(request)
    if request.user.is_authenticated:
        return RedirectResponse(url=next_url)
    google_client_id = os.environ.get("GOOGLE_CLIENT_ID")
    redirect_uri = str(request.app.url_path_for("auth_post"))
    return templates.TemplateResponse(
        "login.html",
        context={
            "request": request,
            "google_client_id": google_client_id,
            "redirect_uri": f"{redirect_uri}?next={next_url}",
        },
    )


@web_client.get("/agents", response_class=HTMLResponse)
def agents_page(request: Request):
    return templates.TemplateResponse(
        "agents/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/settings", response_class=HTMLResponse)
@requires(["authenticated"], redirect="login_page")
def config_page(request: Request):
    return templates.TemplateResponse("settings/index.html", context={"request": request})


@web_client.get("/settings/content/github", response_class=HTMLResponse)
@requires(["authenticated"], redirect="login_page")
def github_config_page(request: Request):
    user = request.user.object
    user_config = get_user_config(user, request)
    current_github_config = get_user_github_config(user)

    if current_github_config:
        raw_repos = current_github_config.githubrepoconfig.all()
        repos = []
        for repo in raw_repos:
            repos.append(
                GithubRepoConfig(
                    name=repo.name,
                    owner=repo.owner,
                    branch=repo.branch,
                )
            )
        current_config = GithubContentConfig(
            pat_token=current_github_config.pat_token,
            repos=repos,
        )
        current_config = json.loads(current_config.json())
    else:
        current_config = {}  # type: ignore

    user_config["current_config"] = current_config
    return templates.TemplateResponse("content_source_github_input.html", context=user_config)


@web_client.get("/share/chat/{public_conversation_slug}", response_class=HTMLResponse)
def view_public_conversation(request: Request):
    return templates.TemplateResponse(
        "share/chat/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/automations", response_class=HTMLResponse)
def automations_config_page(
    request: Request,
):
    return templates.TemplateResponse(
        "automations/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/.well-known/assetlinks.json", response_class=FileResponse)
def assetlinks(request: Request):
    return FileResponse(constants.assetlinks_file_path)


@web_client.get("/server/error", response_class=HTMLResponse)
def server_error_page(request: Request):
    return templates.TemplateResponse("error.html", context={"request": request})



import os
import threading
from collections import defaultdict
from pathlib import Path
from typing import Dict, List

from apscheduler.schedulers.background import BackgroundScheduler
from openai import OpenAI
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from whisper import Whisper

from khoj.database.models import ProcessLock
from khoj.processor.embeddings import CrossEncoderModel, EmbeddingsModel
from khoj.utils import config as utils_config
from khoj.utils.helpers import LRU, get_device, is_env_var_true

# Application Global State
embeddings_model: Dict[str, EmbeddingsModel] = None
cross_encoder_model: Dict[str, CrossEncoderModel] = None
openai_client: OpenAI = None
whisper_model: Whisper = None
log_file: Path = None
verbose: int = 0
host: str = None
port: int = None
ssl_config: Dict[str, str] = None
cli_args: List[str] = None
query_cache: Dict[str, LRU] = defaultdict(LRU)
chat_lock = threading.Lock()
SearchType = utils_config.SearchType
scheduler: BackgroundScheduler = None
schedule_leader_process_lock: ProcessLock = None
telemetry: List[Dict[str, str]] = []
telemetry_disabled: bool = is_env_var_true("KHOJ_TELEMETRY_DISABLE")
khoj_version: str = None
device = get_device()
anonymous_mode: bool = False
pretrained_tokenizers: Dict[str, PreTrainedTokenizer | PreTrainedTokenizerFast] = dict()
billing_enabled: bool = (
    os.getenv("STRIPE_API_KEY") is not None
    and os.getenv("STRIPE_SIGNING_SECRET") is not None
    and os.getenv("KHOJ_CLOUD_SUBSCRIPTION_URL") is not None
)



# Generated by Django 4.2.11 on 2024-06-21 04:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0046_khojuser_email_verification_code_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="entry",
            name="file_type",
            field=models.CharField(
                choices=[
                    ("image", "Image"),
                    ("pdf", "Pdf"),
                    ("plaintext", "Plaintext"),
                    ("markdown", "Markdown"),
                    ("org", "Org"),
                    ("notion", "Notion"),
                    ("github", "Github"),
                    ("conversation", "Conversation"),
                    ("docx", "Docx"),
                ],
                default="plaintext",
                max_length=30,
            ),
        ),
    ]



# Generated by Django 4.2.5 on 2023-10-26 23:52

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0009_khojapiuser"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="Embeddings",
            new_name="Entry",
        ),
        migrations.RenameModel(
            old_name="EmbeddingsDates",
            new_name="EntryDates",
        ),
        migrations.RenameField(
            model_name="entrydates",
            old_name="embeddings",
            new_name="entry",
        ),
        migrations.RenameIndex(
            model_name="entrydates",
            new_name="database_en_date_8d823c_idx",
            old_name="database_em_date_a1ba47_idx",
        ),
    ]



import logging
from datetime import datetime
from typing import Any, AsyncGenerator, Dict, List, Optional

from khoj.database.models import Agent, ChatMessageModel, ChatModel
from khoj.processor.conversation import prompts
from khoj.processor.conversation.openai.utils import (
    chat_completion_with_backoff,
    clean_response_schema,
    completion_with_backoff,
    get_structured_output_support,
    is_openai_api,
    responses_chat_completion_with_backoff,
    responses_completion_with_backoff,
    to_openai_tools,
)
from khoj.processor.conversation.utils import (
    OperatorRun,
    ResponseWithThought,
    StructuredOutputSupport,
    generate_chatml_messages_with_context,
    messages_to_print,
)
from khoj.utils.helpers import ToolDefinition, is_none_or_empty, truncate_code_context
from khoj.utils.rawconfig import FileAttachment, LocationData
from khoj.utils.yaml import yaml_dump

logger = logging.getLogger(__name__)


def send_message_to_model(
    messages,
    api_key,
    model,
    response_type="text",
    response_schema=None,
    tools: list[ToolDefinition] = None,
    deepthought=False,
    api_base_url=None,
    tracer: dict = {},
):
    """
    Send message to model
    """

    model_kwargs: Dict[str, Any] = {}
    json_support = get_structured_output_support(model, api_base_url)
    if tools and json_support == StructuredOutputSupport.TOOL:
        model_kwargs["tools"] = to_openai_tools(tools, use_responses_api=is_openai_api(api_base_url))
    elif response_schema and json_support >= StructuredOutputSupport.SCHEMA:
        # Drop unsupported fields from schema passed to OpenAI APi
        cleaned_response_schema = clean_response_schema(response_schema)
        if is_openai_api(api_base_url):
            model_kwargs["text"] = {
                "format": {
                    "type": "json_schema",
                    "strict": True,
                    "name": response_schema.__name__,
                    "schema": cleaned_response_schema,
                }
            }
        else:
            model_kwargs["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "schema": cleaned_response_schema,
                    "name": response_schema.__name__,
                    "strict": True,
                },
            }
    elif response_type == "json_object" and json_support == StructuredOutputSupport.OBJECT:
        model_kwargs["response_format"] = {"type": response_type}

    # Get Response from GPT
    if is_openai_api(api_base_url):
        return responses_completion_with_backoff(
            messages=messages,
            model_name=model,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            model_kwargs=model_kwargs,
            tracer=tracer,
        )
    else:
        return completion_with_backoff(
            messages=messages,
            model_name=model,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            model_kwargs=model_kwargs,
            tracer=tracer,
        )


async def converse_openai(
    # Query
    user_query: str,
    # Context
    references: list[dict],
    online_results: Optional[Dict[str, Dict]] = None,
    code_results: Optional[Dict[str, Dict]] = None,
    operator_results: Optional[List[OperatorRun]] = None,
    query_images: Optional[list[str]] = None,
    query_files: str = None,
    generated_files: List[FileAttachment] = None,
    generated_asset_results: Dict[str, Dict] = {},
    program_execution_context: List[str] = None,
    location_data: LocationData = None,
    chat_history: list[ChatMessageModel] = [],
    model: str = "gpt-4.1-mini",
    api_key: Optional[str] = None,
    api_base_url: Optional[str] = None,
    temperature: float = 0.6,
    max_prompt_size=None,
    tokenizer_name=None,
    user_name: str = None,
    agent: Agent = None,
    vision_available: bool = False,
    deepthought: Optional[bool] = False,
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    """
    Converse with user using OpenAI's ChatGPT
    """
    # Initialize Variables
    current_date = datetime.now()

    if agent and agent.personality:
        system_prompt = prompts.custom_personality.format(
            name=agent.name,
            bio=agent.personality,
            current_date=current_date.strftime("%Y-%m-%d"),
            day_of_week=current_date.strftime("%A"),
        )
    else:
        system_prompt = prompts.personality.format(
            current_date=current_date.strftime("%Y-%m-%d"),
            day_of_week=current_date.strftime("%A"),
        )

    if location_data:
        location_prompt = prompts.user_location.format(location=f"{location_data}")
        system_prompt = f"{system_prompt}\n{location_prompt}"

    if user_name:
        user_name_prompt = prompts.user_name.format(name=user_name)
        system_prompt = f"{system_prompt}\n{user_name_prompt}"

    context_message = ""
    if not is_none_or_empty(references):
        context_message = f"{prompts.notes_conversation.format(references=yaml_dump(references))}\n\n"
    if not is_none_or_empty(online_results):
        context_message += f"{prompts.online_search_conversation.format(online_results=yaml_dump(online_results))}\n\n"
    if not is_none_or_empty(code_results):
        context_message += (
            f"{prompts.code_executed_context.format(code_results=truncate_code_context(code_results))}\n\n"
        )
    if not is_none_or_empty(operator_results):
        operator_content = [
            {"query": oc.query, "response": oc.response, "webpages": oc.webpages} for oc in operator_results
        ]
        context_message += (
            f"{prompts.operator_execution_context.format(operator_results=yaml_dump(operator_content))}\n\n"
        )

    context_message = context_message.strip()

    # Setup Prompt with Primer or Conversation History
    messages = generate_chatml_messages_with_context(
        user_query,
        system_prompt,
        chat_history,
        context_message=context_message,
        model_name=model,
        max_prompt_size=max_prompt_size,
        tokenizer_name=tokenizer_name,
        query_images=query_images,
        vision_enabled=vision_available,
        model_type=ChatModel.ModelType.OPENAI,
        query_files=query_files,
        generated_files=generated_files,
        generated_asset_results=generated_asset_results,
        program_execution_context=program_execution_context,
    )
    logger.debug(f"Conversation Context for GPT: {messages_to_print(messages)}")

    # Get Response from GPT
    if is_openai_api(api_base_url):
        async for chunk in responses_chat_completion_with_backoff(
            messages=messages,
            model_name=model,
            temperature=temperature,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            tracer=tracer,
        ):
            yield chunk
    else:
        # For non-OpenAI APIs, use the chat completion method
        async for chunk in chat_completion_with_backoff(
            messages=messages,
            model_name=model,
            temperature=temperature,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            tracer=tracer,
        ):
            yield chunk



# Generated by Django 5.1.9 on 2025-06-04 01:11

import uuid

from django.db import migrations, models


def fix_malformed_uuids(apps, schema_editor):
    KhojUser = apps.get_model("database", "KhojUser")

    # Track UUID changes for automation cleanup
    uuid_mappings = {}

    # Handle null or empty user UUIDs
    for user in KhojUser.objects.filter(uuid__isnull=True):
        old_uuid = str(user.uuid) if user.uuid else "None"
        user.uuid = uuid.uuid4()
        user.save()
        uuid_mappings[old_uuid] = str(user.uuid)

    # Handle malformed user UUIDs
    for user in KhojUser.objects.all():
        current_uuid_val = user.uuid
        try:
            if not isinstance(current_uuid_val, uuid.UUID):
                # Attempt to parse it as UUID. This will catch "None", "null" strings or other malformed hex.
                uuid.UUID(str(current_uuid_val))
        except (ValueError, TypeError, AttributeError):
            old_uuid_str = str(current_uuid_val)
            new_uuid_obj = uuid.uuid4()
            user.uuid = new_uuid_obj
            user.save(
                update_fields=["uuid"]
            )  # Important to use update_fields to avoid triggering full save logic if not needed
            uuid_mappings[old_uuid_str] = str(new_uuid_obj)
            print(f"Fixed malformed UUID for user (old: '{old_uuid_str}', new: {str(new_uuid_obj)})")

    # Clean up orphaned automations
    cleanup_orphaned_automations(uuid_mappings)


def cleanup_orphaned_automations(uuid_mappings):
    """Remove automations with malformed UUIDs in job_ids"""
    from apscheduler.jobstores.base import JobLookupError

    from khoj.utils import state

    if not state.scheduler:
        return

    all_jobs = state.scheduler.get_jobs()
    removed_orphaned_count = 0
    removed_malformed_count = 0

    for job in all_jobs:
        if job.id.startswith("automation_"):
            # Extract UUID from job_id: "automation_{uuid}_{query_id}"
            job_parts = job.id.split("_", 2)
            if len(job_parts) >= 2:
                job_uuid = job_parts[1]

                # Check if this UUID was malformed
                if job_uuid in uuid_mappings:
                    # Remove orphaned automation
                    try:
                        state.scheduler.remove_job(job.id)
                        removed_orphaned_count += 1
                        print(f"Removed orphaned automation: {job.id}")
                    except JobLookupError:
                        pass  # Job already removed

                # Also remove jobs with clearly malformed UUIDs
                elif job_uuid in ["None", "null"] or not is_valid_uuid(job_uuid):
                    try:
                        state.scheduler.remove_job(job.id)
                        removed_malformed_count += 1
                        print(f"Removed automation with malformed UUID: {job.id}")
                    except JobLookupError:
                        pass

    if removed_orphaned_count > 0 or removed_malformed_count > 0:
        print(f"Removed {removed_orphaned_count} orphaned and {removed_malformed_count} malformed automations.")


def is_valid_uuid(uuid_string):
    """Check if string is a valid UUID"""
    try:
        uuid.UUID(str(uuid_string))
        return True
    except (ValueError, TypeError, AttributeError):
        return False


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0089_chatmodel_price_tier_and_more"),
    ]

    operations = [
        migrations.RunPython(fix_malformed_uuids, reverse_code=migrations.RunPython.noop),
        migrations.AlterField(
            model_name="khojuser",
            name="uuid",
            field=models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
        ),
    ]
